#include "dsStateMgr.h"
#include "../dsLog.h"

enum TokenType {
    kTokNone, kTokName, kTokSymbol, kTokNumber, kTokString, kTokComment
};

struct Token {
    TokenType fType;
    dsString fText;
    long fLine;
};

static inline int peekch(FILE* f)
{
    int ch = fgetc(f);
    if (ch != EOF)
        ungetc(ch, f);
    return ch;
}

static std::list<Token> tokenize(const dsString& filename)
{
    TokenType state = kTokNone;
    char* psave = NULL;
    char* it;
    dsString tempTok;
    std::list<Token> tokens;

    FILE* fin = fopen(filename.cstr(), "rb");
    char buf[4096];
    long line = 1;
    bool haveEol = false;
    while (peekch(fin) != EOF) {
        tempTok = (psave != NULL) ? tempTok + psave : "";
        if (haveEol) {
            ++line;
            haveEol = false;
        }
        fgets(buf, 4096, fin);
        it = buf;
        while (*it != 0) {
            if (*it == '#') {
                if (state != kTokString)
                    state = kTokComment;
            } else if ((*it >= 'A' && *it <= 'Z') || (*it >= 'a' && *it <= 'z')
                       || (*it == '$')) {
                if (state == kTokNone) {
                    state = kTokName;
                    psave = it;
                } else if (state == kTokNumber) {
                    Token tok;
                    tok.fType = state;
                    tok.fText = tempTok + dsString(psave, it - psave);
                    tok.fLine = line;
                    tokens.push_back(tok);
                    state = kTokName;
                    psave = it;
                }
            } else if (*it >= '0' && *it <= '9') {
                if (state == kTokNone) {
                    state = kTokNumber;
                    psave = it;
                }
            } else if (*it == '{' || *it == '}' || *it == '[' || *it == ']'
                       || *it == '(' || *it == ')' || *it == '=' || *it == ','
                       || *it == '.') {
                if (state == kTokName || state == kTokNumber) {
                    Token tok;
                    tok.fType = state;
                    tok.fText = tempTok + dsString(psave, it - psave);
                    tok.fLine = line;
                    tokens.push_back(tok);
                    tok.fType = kTokSymbol;
                    tok.fText = dsString(it, 1);
                    tok.fLine = line;
                    tokens.push_back(tok);
                    state = kTokNone;
                    psave = NULL;
                } else if (state == kTokNone) {
                    Token tok;
                    tok.fType = kTokSymbol;
                    tok.fText = dsString(it, 1);
                    tok.fLine = line;
                    tokens.push_back(tok);
                }
            } else if (*it == '"') {
                if (state == kTokName || state == kTokNumber) {
                    Token tok;
                    tok.fType = state;
                    tok.fText = tempTok + dsString(psave, it - psave);
                    tok.fLine = line;
                    tokens.push_back(tok);
                    state = kTokString;
                    psave = it + 1;
                } else if (state == kTokNone) {
                    state = kTokString;
                    psave = it + 1;
                } else if (state == kTokString) {
                    Token tok;
                    tok.fType = kTokString;
                    tok.fText = tempTok + dsString(psave, it - psave);
                    tok.fLine = line;
                    tokens.push_back(tok);
                    state = kTokNone;
                    psave = NULL;
                }
            } else if (*it == '\t' || *it == ' ') {
                if (state == kTokName || state == kTokNumber) {
                    Token tok;
                    tok.fType = state;
                    tok.fText = tempTok + dsString(psave, it - psave);
                    tok.fLine = line;
                    tokens.push_back(tok);
                    state = kTokNone;
                    psave = NULL;
                }
            } else if (*it == '\r' || *it == '\n') {
                if (state == kTokName || state == kTokNumber) {
                    Token tok;
                    tok.fType = state;
                    tok.fText = tempTok + dsString(psave, it - psave);
                    tok.fLine = line;
                    tokens.push_back(tok);
                    state = kTokNone;
                    psave = NULL;
                } else if (state == kTokComment) {
                    state = kTokNone;
                }
                haveEol = true;
            } else {
                if (state != kTokComment && state != kTokString) {
                    dsLogF(kLogError, "%s:%ld: Unexpected character '%c'",
                                    filename.cstr().data(), line, *it);
                    state = kTokNone;
                    psave = NULL;
                }
            }
            it++;
        }
    }
    fclose(fin);
    return tokens;
}

static dsVarDescriptor parseVariable(std::list<Token>& file, const dsString& filename,
                                     dsStateDescriptor* state)
{
    dsVarDescriptor var(state);

    if (file.empty()) {
        dsLogF(kLogError, "%s: Unexpected end of file", filename.cstr().data());
        return dsVarDescriptor(NULL);
    }
    Token& tok = file.front();
    if (tok.fType != kTokName) {
        dsLogF(kLogError, "%s:%ld: Invalid variable type '%s'",
                          filename.cstr().data(), tok.fLine,
                          tok.fText.cstr().data());
        return dsVarDescriptor(NULL);
    }
    var.setType(dsVarDescriptor::typeFromName(tok.fText));
    if (var.type() == dsVarDescriptor::kNone) {
        dsLogF(kLogError, "%s:%ld: Invalid variable type '%s'",
                          filename.cstr().data(), tok.fLine,
                          tok.fText.cstr().data());
        return dsVarDescriptor(NULL);
    }

    file.pop_front();
    if (file.empty()) {
        dsLogF(kLogError, "%s: Unexpected end of file", filename.cstr().data());
        return dsVarDescriptor(NULL);
    }
    tok = file.front();
    if (tok.fType != kTokName) {
        dsLogF(kLogError, "%s:%ld: Invalid variable name '%s'",
                          filename.cstr().data(), tok.fLine,
                          tok.fText.cstr().data());
        return dsVarDescriptor(NULL);
    }
    var.setName(tok.fText);

    file.pop_front();
    if (file.empty()) {
        dsLogF(kLogError, "%s: Unexpected end of file", filename.cstr().data());
        return dsVarDescriptor(NULL);
    }
    tok = file.front();
    if (tok.fType != kTokSymbol || tok.fText != "[") {
        dsLogF(kLogError, "%s:%ld: Unexpected '%s', expected '['",
                          filename.cstr().data(), tok.fLine,
                          tok.fText.cstr().data());
        return dsVarDescriptor(NULL);
    }

    file.pop_front();
    if (file.empty()) {
        dsLogF(kLogError, "%s: Unexpected end of file", filename.cstr().data());
        return dsVarDescriptor(NULL);
    }
    tok = file.front();
    if (tok.fType == kTokSymbol && tok.fText == "]") {
        var.setCount(-1);
    } else if (tok.fType == kTokNumber) {
        var.setCount(tok.fText.toInt());
        file.pop_front();
        if (file.empty()) {
            dsLogF(kLogError, "%s: Unexpected end of file", filename.cstr().data());
            return dsVarDescriptor(NULL);
        }
        tok = file.front();
        if (tok.fType != kTokSymbol || tok.fText != "]") {
            dsLogF(kLogError, "%s:%ld: Unexpected '%s', expected ']'",
                              filename.cstr().data(), tok.fLine,
                              tok.fText.cstr().data());
            return dsVarDescriptor(NULL);
        }
    } else {
        dsLogF(kLogError, "%s:%ld: Unexpected '%s', expected NUMBER or ']'",
                          filename.cstr().data(), tok.fLine,
                          tok.fText.cstr().data());
        return dsVarDescriptor(NULL);
    }

    while (true) {
        file.pop_front();
        if (file.empty())
            break;
        tok = file.front();
        if (tok.fType != kTokName) {
            // Not a recognized token -- finish the variable
            break;
        } else if (tok.fText == "DEFAULT") {
            file.pop_front();
            if (file.empty()) {
                dsLogF(kLogError, "%s: Unexpected end of file", filename.cstr().data());
                return dsVarDescriptor(NULL);
            }
            tok = file.front();
            if (tok.fType != kTokSymbol || tok.fText != "=") {
                dsLogF(kLogError, "%s:%ld: Unexpected '%s', expected '='",
                                  filename.cstr().data(), tok.fLine,
                                  tok.fText.cstr().data());
                return var;
            }
            int nParens = 0;
            dsString def;
            do {
                file.pop_front();
                if (file.empty()) {
                    dsLogF(kLogError, "%s: Unexpected end of file", filename.cstr().data());
                    return dsVarDescriptor(NULL);
                }
                tok = file.front();
                def += tok.fText;
                if (tok.fType == kTokSymbol && tok.fText == "(")
                    nParens++;
                else if (tok.fType == kTokSymbol && tok.fText == ")")
                    nParens--;
            } while (nParens > 0);
        } else if (tok.fText == "DEFAULTOPTION") {
            file.pop_front();
            if (file.empty()) {
                dsLogF(kLogError, "%s: Unexpected end of file", filename.cstr().data());
                return dsVarDescriptor(NULL);
            }
            tok = file.front();
            if (tok.fType != kTokSymbol || tok.fText != "=") {
                dsLogF(kLogError, "%s:%ld: Unexpected '%s', expected '='",
                                  filename.cstr().data(), tok.fLine,
                                  tok.fText.cstr().data());
                return var;
            }
            file.pop_front();
            if (file.empty()) {
                dsLogF(kLogError, "%s: Unexpected end of file", filename.cstr().data());
                return dsVarDescriptor(NULL);
            }
            tok = file.front();
            if (tok.fType != kTokName) {
                dsLogF(kLogError, "%s:%ld: Unexpected '%s', expected 'VAULT' or 'HIDDEN'",
                                  filename.cstr().data(), tok.fLine,
                                  tok.fText.cstr().data());
                return var;
            }
            if (tok.fText == "VAULT") {
                var.setAlwaysNew();
            } else if (tok.fText == "HIDDEN") {
                var.setInternal();
            } else {
                dsLogF(kLogError, "%s:%ld: Unexpected '%s', expected 'VAULT' or 'HIDDEN'",
                                  filename.cstr().data(), tok.fLine,
                                  tok.fText.cstr().data());
                return var;
            }
        } else if (tok.fText == "DISPLAYOPTION") {
            file.pop_front();
            if (file.empty()) {
                dsLogF(kLogError, "%s: Unexpected end of file", filename.cstr().data());
                return dsVarDescriptor(NULL);
            }
            tok = file.front();
            if (tok.fType != kTokSymbol || tok.fText != "=") {
                dsLogF(kLogError, "%s:%ld: Unexpected '%s', expected '='",
                                  filename.cstr().data(), tok.fLine,
                                  tok.fText.cstr().data());
                return var;
            }
            file.pop_front();
            if (file.empty()) {
                dsLogF(kLogError, "%s: Unexpected end of file", filename.cstr().data());
                return dsVarDescriptor(NULL);
            }
            tok = file.front();
            if (tok.fType != kTokName) {
                dsLogF(kLogError, "%s:%ld: Unexpected '%s', expected IDENTIFIER",
                                  filename.cstr().data(), tok.fLine,
                                  tok.fText.cstr().data());
                return var;
            }
            var.setDisplay(tok.fText);
        } else {
            // Not a recognized keyword -- finish the variable
            break;
        }
    }
    return var;
}

static dsStateDescriptor parseDescriptor(std::list<Token>& file, const dsString& filename)
{
    dsStateDescriptor state;

    if (file.empty())
        // No more descriptors in this file
        return state;

    Token& tok = file.front();
    if (tok.fType != kTokName || tok.fText != "STATEDESC") {
        dsLogF(kLogError, "%s:%ld: Unexpected '%s', expected 'STATEDESC'",
                          filename.cstr().data(), tok.fLine,
                          tok.fText.cstr().data());
        return state;
    }

    file.pop_front();
    if (file.empty()) {
        dsLogF(kLogError, "%s: Unexpected end of file", filename.cstr().data());
        return state;
    }
    tok = file.front();
    if (tok.fType != kTokSymbol || tok.fText != "{") {
        dsLogF(kLogError, "%s:%ld: Unexpected '%s', expected '{'",
                          filename.cstr().data(), tok.fLine,
                          tok.fText.cstr().data());
        return state;
    }

    while (true) {
        file.pop_front();
        if (file.empty()) {
            dsLogF(kLogError, "%s: Unexpected end of file", filename.cstr().data());
            return state;
        }
        tok = file.front();

        if (tok.fType == kTokSymbol && tok.fText == "}") {
            if (state.version() < 0) {
                dsLogF(kLogWarning, "%s: No VERSION set for '%s', using 0",
                                    state.name().cstr().data());
                state.setVersion(0);
            }
            return state;
        }
        if (tok.fType != kTokName) {
            dsLogF(kLogError, "%s:%ld: Unexpected '%s', expected 'VAR' or 'VERSION'",
                              filename.cstr().data(), tok.fLine,
                              tok.fText.cstr().data());
        } else if (tok.fText == "VAR") {
            file.pop_front();
            dsVarDescriptor var = parseVariable(file, filename, &state);
            if (var.isValid())
                state.addVariable(var);
        } else if (tok.fText == "VERSION") {
            file.pop_front();
            if (file.empty()) {
                dsLogF(kLogError, "%s: Unexpected end of file", filename.cstr().data());
                return state;
            }
            tok = file.front();
            if (tok.fType != kTokNumber) {
                dsLogF(kLogError, "%s:%ld: Unexpected '%s', expected NUMBER",
                                  filename.cstr().data(), tok.fLine,
                                  tok.fText.cstr().data());
                // Don't break
            } else {
                state.setVersion(tok.fText.toInt());
            }
        } else {
            dsLogF(kLogError, "%s:%ld: Unexpected '%s', expected 'VAR' or 'VERSION'",
                              filename.cstr().data(), tok.fLine,
                              tok.fText.cstr().data());
            // Don't break
        }
    }
    // We should never get here
    return state;
}

void dsStateMgr::readDescriptors(const dsString& path)
{
    //
}

dsStateDescriptor dsStateMgr::get(const dsString& name, dsInt32 version) const
{
    std::map<dsString, std::list<dsStateDescriptor> >::const_iterator f = fDescriptors.find(name);
    if (f == fDescriptors.end() || f->second.size() == 0)
        return dsStateDescriptor();

    dsStateDescriptor high = f->second.front();
    std::list<dsStateDescriptor>::const_iterator it;
    for (it = f->second.begin(); it != f->second.end(); it++) {
        if (version >= 0 && it->version() == version)
            return *it;
        if (it->version() > high.version())
            high = *it;
    }
    if (version < 0)
        return high;
    return dsStateDescriptor();
}
