#include "dsAuthServer.h"
#include "../dsConfigMgr.h"
#include "../dsLog.h"
#include "../Connection.h"
#include <unistd.h>
#include <cstdlib>
#include <errno.h>

#include "AuthMessages.h"
#include "dsAuthData.h"
#include "../vault/dsVaultNode.h"

static dsUInt32 s_droidKey[4] = { 0, 0, 0, 0 };

static dsUInt32 getInternalTransId()
{
    static dsUInt32 sTransId = 0x80000000;
    return ++sTransId;
}

pid_t dsAuthServer::Create(dsSocket* sock)
{
    pid_t pid = fork();
    if (pid < 0) {
        dsLogF(kLogError, "Error spawning Auth server: %s", strerror(errno));
        delete sock;
        exit(1);
    } else if (pid == 0) {
        s_listenSock.close();
        s_children.clear();
        dsString logpath = dsConfigMgr::GetPath("LogDir", "AuthServer");
        dsLogProcessor* authLog = dsLogProcessor::Create(logpath.cstr(),
                                  (ELogType)dsConfigMgr::GetUInt("LogLevel", LOGLEVEL_DEFAULT));
        authLog->setAddress(sock->getRemoteIpStr().cstr());
        dsLogProcessor::setLog(authLog, true);

        dsAuthServer server(sock);
        server.run();
        exit(0);
    } else {
        // Get rid of this socket and go back to listening
        delete sock;
        return pid;
    }
}

void dsAuthServer::run()
{
    dsLog(kLogDebug, "Setting up encryption parameters");

    // Initialize the connection encryption
    AuthConnectHeader connectInfo;
    fSock.peek(&connectInfo.fHeaderSize, sizeof(dsUInt32));
    if (connectInfo.fHeaderSize != sizeof(connectInfo)) {
        dsLog(kLogError, "Invalid Auth Server connection packet");
        return;
    }
    if (fSock.recv(&connectInfo, sizeof(connectInfo)) != sizeof(connectInfo)) {
        dsLog(kLogError, "Invalid Auth Server connection packet");
        return;
    }

    NetCli_Cli2Srv_Connect cryptConn;
    if (fSock.recv(&cryptConn.fMessage, sizeof(dsUByte)) != sizeof(dsUByte)
        || cryptConn.fMessage != kNetCliCli2SrvConnect) {
        dsLog(kLogError, "Invalid NetCli connection packet");
        return;
    }
    if (fSock.recv(&cryptConn.fLength, sizeof(dsUByte)) != sizeof(dsUByte)) {
        dsLog(kLogError, "Invalid NetCli connection packet");
        return;
    }

    size_t keyDataLen = cryptConn.fLength - (sizeof(dsUByte) * 2);
    if (keyDataLen > 64) {
        dsLog(kLogError, "Seed value too long");
        return;
    }
    memset(cryptConn.fKeyYData, 0, 64);
    fSock.recv(cryptConn.fKeyYData, keyDataLen);

    NetCli_Srv2Cli_Encrypt cryptReply;
    cryptReply.fMessage = kNetCliSrv2CliEncrypt;
    cryptReply.fLength = sizeof(cryptReply);
    if (!calculateKeys(dsBigInteger(cryptConn.fKeyYData, keyDataLen))) {
        dsLog(kLogError, "Error setting up encryption keys");
        return;
    }
    memcpy(cryptReply.fSeed, fServerSeed, 7);
    fSock.send(&cryptReply, sizeof(cryptReply));
    fSock.init(7, fSharedKey);

    // Acquire the auth database:
    dsString host = dsConfigMgr::GetString("dbhost", DBHOST_DEFAULT);
    dsString port = dsConfigMgr::GetString("dbport", DBPORT_DEFAULT);
    dsString user = dsConfigMgr::GetString("dbuser", DBUSER_DEFAULT);
    dsString pass = dsConfigMgr::GetString("dbpass", DBPASS_DEFAULT);
    dsString sslMode = dsConfigMgr::GetString("dbsslmode", DBSSLMODE_DEFAULT);
    dsString dbname = dsConfigMgr::GetString("dbname", DBNAME_DEFAULT);
    if (!fAuthDb.connect(host.cstr(), port.cstr(), user.cstr(), pass.cstr(),
                         sslMode.cstr(), dbname.cstr())) {
        dsLogF(kLogError, "Could not connect to Auth Database: %s",
                          fAuthDb.getErrorMsg());
        fAuthDb.close();
        return;
    }
    dsLog(kLogInfo, "Auth server started");

    // Send game server address:
    msgparm_t* replyBuf;
    {
        const dsNetMsg* lobbyServerMsg = GET_Auth2Cli(kAuth2Cli_ServerAddr);
        replyBuf = AllocMessage(lobbyServerMsg);
        replyBuf[0] = dsSocket::GetAddress(dsConfigMgr::GetString("GameHost").cstr());
        dsUuid nullUuid;
        memcpy((void*)replyBuf[1], &nullUuid, sizeof(dsUuid));
        fSock.sendMsg(replyBuf, lobbyServerMsg);
        FreeMessage(replyBuf, lobbyServerMsg);
    }

    // Seed the RNG for use in some messages below:
    srandom((unsigned int)time(NULL));

    dsUInt16 msgType;
    while (fSock.recv(&msgType, sizeof(dsUInt16)) > 0) {
        const dsNetMsg* msg = GET_Cli2Auth(msgType);
        msgparm_t* reqBuf = fSock.recvMsg(msg);

        switch (msgType) {
        case kCli2Auth_PingRequest:
            {
                const dsNetMsg* reply = GET_Auth2Cli(kAuth2Cli_PingReply);
                replyBuf = AllocMessage(reply);
                replyBuf[0] = reqBuf[0];    // Ping Time
                replyBuf[1] = reqBuf[1];    // Trans ID
                replyBuf[2] = reqBuf[2];    // Payload size
                replyBuf[3] = reqBuf[3];    // Payload data
                fSock.sendMsg(replyBuf, reply);

                replyBuf[3] = 0;    // Don't delete twice!
                FreeMessage(replyBuf, reply);
                dsLog(kLogDebug, "PING!");
            }
            break;
        case kCli2Auth_ClientRegisterRequest:
            {
                const dsNetMsg* reply = GET_Auth2Cli(kAuth2Cli_ClientRegisterReply);
                replyBuf = AllocMessage(reply);
                replyBuf[0] = 0;    // Server Challenge
                fSock.sendMsg(replyBuf, reply);

                FreeMessage(replyBuf, reply);
                dsLog(kLogDebug, "Client Registered");
            }
            break;
        //case kCli2Auth_ClientSetCCRLevel:
        case kCli2Auth_AcctLoginRequest:
            {
                dsString acctName = (dsChar*)(reqBuf[2]);
                dsSha1Hash passHash = *(dsSha1Hash*)(reqBuf[3]);
                dsString authToken = (dsChar*)(reqBuf[4]);
                dsString os = (dsChar*)(reqBuf[5]);
                dsLogF(kLogInfo, "[Login] U:%s P:%s T:%s OS:%s",
                                 acctName.cstr().data(),
                                 passHash.toString().cstr().data(),
                                 authToken.cstr().data(),
                                 os.cstr().data());
                login(acctName.cstr(), passHash);
                if (fLoginInfo.fResult != kNetSuccess)
                    dsLogF(kLogInfo, "[Login] Login failed: %d", fLoginInfo.fResult);
                else
                    dsLogF(kLogInfo, "[Login] ID: %s", fLoginInfo.fAcctUuid.asString().cstr().data());

                // Get the players assigned to this account and send them
                // to the client:
                dsPostgresResult res = fAuthDb.query(dsString::Format(
                    "SELECT * FROM auth.\"Players\" WHERE \"AcctUuid\"='%s'",
                    fLoginInfo.fAcctUuid.asString().cstr().data()));
                if (!res.isValid()) {
                    dsLogF(kLogError, "SQL Error getting players: %s", fAuthDb.getErrorMsg());
                    fLoginInfo.fResult = kNetErrInternalError;
                } else {
                    const dsNetMsg* reply = GET_Auth2Cli(kAuth2Cli_AcctPlayerInfo);
                    for (size_t i=0; i<res.getNumRows(); i++) {
                        dsPlayerInfo plrInfo;
                        dsPostgresRow row = res.fetchRow();
                        plrInfo.fPlayerIdx = row["\"PlayerIdx\""].getUInt();
                        plrInfo.fExplorer = row["\"Explorer\""].getUInt();
                        plrInfo.fPlayerName = row["\"PlayerName\""].getString();
                        plrInfo.fAvatarShape = row["\"AvatarShape\""].getString();

                        replyBuf = AllocMessage(reply);
                        replyBuf[0] = reqBuf[0];    // Trans ID
                        replyBuf[1] = plrInfo.fPlayerIdx;
                        replyBuf[2] = (msgparm_t)plrInfo.fPlayerName.data();
                        replyBuf[3] = (msgparm_t)plrInfo.fAvatarShape.data();
                        replyBuf[4] = plrInfo.fExplorer;
                        fSock.sendMsg(replyBuf, reply);

                        // Don't delete these
                        replyBuf[2] = 0;
                        replyBuf[3] = 0;
                        FreeMessage(replyBuf, reply);
                    }
                }

                // Final login message, after all players are sent:
                const dsNetMsg* reply = GET_Auth2Cli(kAuth2Cli_AcctLoginReply);
                replyBuf = AllocMessage(reply);
                replyBuf[0] = reqBuf[0];    // Trans ID
                replyBuf[1] = fLoginInfo.fResult;
                memcpy((void*)(replyBuf[2]), &fLoginInfo.fAcctUuid, sizeof(dsUuid));
                replyBuf[3] = fLoginInfo.fAcctFlags;
                replyBuf[4] = fLoginInfo.fBillingType;
                memcpy((void*)(replyBuf[5]), s_droidKey, sizeof(s_droidKey));
                fSock.sendMsg(replyBuf, reply);
                FreeMessage(replyBuf, reply);
            }
            break;
        //case kCli2Auth_AcctSetEulaVersion:
        //case kCli2Auth_AcctSetDataRequest:
        case kCli2Auth_AcctSetPlayerRequest:
            {
                fPlayerInfo.fResult = kNetPending;
                if (reqBuf[1] > 0) {
                    dsLogF(kLogInfo, "[Auth] Requesting Player ID %llu", reqBuf[1]);
                    dsPostgresResult res = fAuthDb.query(dsString::Format(
                        "SELECT * FROM auth.\"Players\" WHERE \"AcctUuid\"='%s' "
                                                       "AND \"PlayerIdx\"=%llu",
                        fLoginInfo.fAcctUuid.asString().cstr().data(),
                        reqBuf[1]));
                    if (res.getNumRows() == 0) {
                        dsLogF(kLogWarning, "[Auth] Player ID %llu not available!", reqBuf[1]);
                        fPlayerInfo.fResult = kNetErrPlayerNotFound;
                    } else {
                        dsPostgresRow row = res.fetchRow();
                        fPlayerInfo.fPlayerIdx = row["\"PlayerIdx\""].getUInt();
                        fPlayerInfo.fExplorer = row["\"Explorer\""].getUInt();
                        fPlayerInfo.fPlayerName = row["\"PlayerName\""].getString();
                        fPlayerInfo.fAvatarShape = row["\"AvatarShape\""].getString();
                        fPlayerInfo.fResult = kNetSuccess;
                        dsLog(kLogDebug, "[Auth] Successfully got Player Info");
                    }
                } else {
                    dsLog(kLogDebug, "[Auth] Requesting No Player");
                    fPlayerInfo.fPlayerIdx = 0;
                    fPlayerInfo.fResult = kNetSuccess;
                }

                const dsNetMsg* reply = GET_Auth2Cli(kAuth2Cli_AcctSetPlayerReply);
                replyBuf = AllocMessage(reply);
                replyBuf[0] = reqBuf[0];    // Trans ID
                replyBuf[1] = fPlayerInfo.fResult;
                fSock.sendMsg(replyBuf, reply);
                FreeMessage(replyBuf, reply);
            }
            break;
        //case kCli2Auth_AcctCreateRequest:
        //case kCli2Auth_AcctChangePasswordRequest:
        //case kCli2Auth_AcctSetRolesRequest:
        //case kCli2Auth_AcctSetBillingTypeRequest:
        //case kCli2Auth_AcctActivateRequest:
        //case kCli2Auth_AcctCreateFromKeyRequest:
        //case kCli2Auth_PlayerDeleteRequest:
        //case kCli2Auth_PlayerUndeleteRequest:
        //case kCli2Auth_PlayerSelectRequest:
        //case kCli2Auth_PlayerRenameRequest:
        case kCli2Auth_PlayerCreateRequest:
            {
                const dsNetMsg* reply = GET_Auth2Cli(kAuth2Cli_PlayerCreateReply);
                replyBuf = AllocMessage(reply);
                replyBuf[0] = reqBuf[0];    // Trans ID
                replyBuf[2] = 0;            // Player ID
                replyBuf[3] = 0;            // Explorer

                dsPlayerInfo player;
                dsString playerName = (dsChar*)(reqBuf[1]);
                dsString avatarShape = (dsChar*)(reqBuf[2]);
                dsString friendInvite = (dsChar*)(reqBuf[3]);
                if (fLoginInfo.fResult != kNetSuccess) {
                    replyBuf[1] = kNetErrAuthenticationFailed;
                } else {
                    player = createPlayer(playerName.cstr(),
                                          avatarShape.cstr(),
                                          friendInvite.cstr());
                    replyBuf[1] = player.fResult;
                    replyBuf[2] = player.fPlayerIdx;
                    replyBuf[3] = player.fExplorer;
                    replyBuf[4] = (msgparm_t)playerName.data();
                    replyBuf[5] = (msgparm_t)avatarShape.data();
                }
                fSock.sendMsg(replyBuf, reply);

                // Don't delete these
                replyBuf[4] = 0;
                replyBuf[5] = 0;
                FreeMessage(replyBuf, reply);
            }
            break;
        //case kCli2Auth_PlayerSetStatus:
        //case kCli2Auth_PlayerChat:
        //case kCli2Auth_UpgradeVisitorRequest:
        //case kCli2Auth_SetPlayerBanStatusRequest:
        //case kCli2Auth_KickPlayer:
        //case kCli2Auth_ChangePlayerNameRequest:
        //case kCli2Auth_SendFriendInviteRequest:
        case kCli2Auth_VaultNodeCreate:
            {
                dsUInt32 vaultTrans = fVault.sendCreateRequest((dsUByte*)(reqBuf[2]), reqBuf[1]);
                fVaultTransMutex.lock();
                fVaultTransMap[vaultTrans] = reqBuf[0];
                fVaultTransMutex.unlock();
            }
            break;
        case kCli2Auth_VaultNodeFetch:
            {
                dsUInt32 vaultTrans = fVault.sendFetchRequest(reqBuf[1]);
                fVaultTransMutex.lock();
                fVaultTransMap[vaultTrans] = reqBuf[0];
                fVaultTransMutex.unlock();
            }
            break;
        case kCli2Auth_VaultNodeSave:
            {
                dsUInt32 vaultTrans = fVault.sendSaveRequest(reqBuf[1]              /* Node ID */,
                                                             *(dsUuid*)(reqBuf[2])  /* Revision ID */,
                                                             (dsUByte*)(reqBuf[4])  /* Node Data */,
                                                             reqBuf[3]              /* Node Size */);
                fVaultTransMutex.lock();
                fVaultTransMap[vaultTrans] = reqBuf[0];
                fVaultTransMutex.unlock();
            }
            break;
        //case kCli2Auth_VaultNodeDelete:
        case kCli2Auth_VaultNodeAdd:
            {
                dsUInt32 vaultTrans = fVault.sendAddRequest(reqBuf[1]   /* Parent ID */,
                                                            reqBuf[2]   /* Child ID */,
                                                            reqBuf[3]   /* Owner ID */);
                fVaultTransMutex.lock();
                fVaultTransMap[vaultTrans] = reqBuf[0];
                fVaultTransMutex.unlock();
            }
            break;
        //case kCli2Auth_VaultNodeRemove:
        case kCli2Auth_VaultFetchNodeRefs:
            {
                dsUInt32 vaultTrans = fVault.sendFetchRefsRequest(reqBuf[1]);
                fVaultTransMutex.lock();
                fVaultTransMap[vaultTrans] = reqBuf[0];
                fVaultTransMutex.unlock();
            }
            break;
        case kCli2Auth_VaultInitAgeRequest:
            {
                // First Initialize the age
                dsUuid ageInstId = *(dsUuid*)(reqBuf[1]);
                dsUuid parentAgeId = *(dsUuid*)(reqBuf[2]);
                dsString ageFilename = (dsChar*)(reqBuf[3]);
                dsString ageInstName = (dsChar*)(reqBuf[4]);
                dsString ageUserName = (dsChar*)(reqBuf[5]);
                dsString ageDescription = (dsChar*)(reqBuf[6]);

                dsLogF(kLogInfo, "Initializing age U[%s] P[%s]: %s",
                       ageInstId.asString().cstr().data(),
                       parentAgeId.asString().cstr().data(),
                       ageDescription.cstr().data());

                dsAgeInfo ageInfo = initAge(ageInstId, parentAgeId,
                                            ageFilename.cstr(), ageInstName.cstr(),
                                            ageUserName.cstr(), ageDescription.cstr(),
                                            reqBuf[7] /* Age Sequence Number */,
                                            reqBuf[8] /* Age Language */);
                const dsNetMsg* reply = GET_Auth2Cli(kAuth2Cli_VaultInitAgeReply);
                replyBuf = AllocMessage(reply);
                replyBuf[0] = reqBuf[0];    // Trans ID
                replyBuf[1] = ageInfo.fResult;
                replyBuf[2] = ageInfo.fAgeIdx;
                replyBuf[3] = ageInfo.fAgeInfoIdx;
                fSock.sendMsg(replyBuf, reply);
                FreeMessage(replyBuf, reply);

                /*
                // Now that it's inited, we probably want to link, so send the age
                if (ageInfo.fResult == kNetSuccess) {
                    // Ajust the Age Instance UUID (Somebody please tell me WHY OH WHY!?)
                    dsUuid adjustedId = ageInstId;
                    adjustedId.fData4[3] = (adjustedId.fData4[2] << 4)
                                         | (adjustedId.fData4[3] >> 4);

                    reply = GET_Auth2Cli(kAuth2Cli_AgeReply);
                    replyBuf = AllocMessage(reply);
                    replyBuf[0] = reqBuf[0];    // Trans ID
                    replyBuf[1] = kNetSuccess;
                    replyBuf[2] = 0;            // Age MCP ID
                    memcpy((void*)(replyBuf[3]), &adjustedId, sizeof(dsUuid));
                    replyBuf[4] = ageInfo.fAgeIdx;
                    replyBuf[5] = dsSocket::GetAddress(dsConfigMgr::GetString("GameHost").cstr());

                    fSock.sendMsg(replyBuf, reply);
                    FreeMessage(replyBuf, reply);
                }
                */
            }
            break;
        case kCli2Auth_VaultNodeFind:
            {
                dsUInt32 vaultTrans = fVault.sendFindRequest((dsUByte*)(reqBuf[2]), reqBuf[1]);
                fVaultTransMutex.lock();
                fVaultTransMap[vaultTrans] = reqBuf[0];
                fVaultTransMutex.unlock();
            }
            break;
        //case kCli2Auth_VaultSetSeen:
        //case kCli2Auth_VaultSendNode:
        case kCli2Auth_AgeRequest:
            {
                const dsNetMsg* reply = GET_Auth2Cli(kAuth2Cli_AgeReply);
                replyBuf = AllocMessage(reply);
                replyBuf[0] = reqBuf[0];    // Trans ID
                replyBuf[2] = 0;            // Age MCP ID
                replyBuf[5] = dsSocket::GetAddress(dsConfigMgr::GetString("GameHost").cstr());

                dsString ageName = (dsChar*)(reqBuf[1]);
                dsUuid ageUuid = *(dsUuid*)(reqBuf[2]);
                dsAgeInfo info = findAge(ageName.cstr(), ageUuid);
                replyBuf[1] = info.fResult;
                replyBuf[4] = info.fAgeIdx;
                if (info.fResult == kNetSuccess)
                    memcpy((void*)(replyBuf[3]), &info.fAgeInstId, sizeof(dsUuid));
                fSock.sendMsg(replyBuf, reply);
                FreeMessage(replyBuf, reply);
            }
            break;
        case kCli2Auth_FileListRequest:
            {
                dsString directory = (dsChar*)(reqBuf[1]);
                dsString ext = (dsChar*)(reqBuf[2]);

                // Get the manifest and send it
                const dsNetMsg* reply = GET_Auth2Cli(kAuth2Cli_FileListReply);
                replyBuf = AllocMessage(reply);
                replyBuf[0] = reqBuf[0];    // Trans ID

                dsAuthFileList mfs;
                mfs.read(directory, ext);
                replyBuf[1] = mfs.getStatus();
                dsChar* data = NULL;
                dsLogF(kLogDebug, "Sending List for %s\\*.%s",
                                  directory.cstr().data(),
                                  ext.cstr().data());
                if (replyBuf[1] == kNetSuccess) {
                    size_t dataSize = mfs.calcTotalSize();
                    data = new dsChar[dataSize];
                    mfs.write(data, dataSize);
                    replyBuf[1] = mfs.getStatus();  // In case the write() fails
                    replyBuf[2] = dataSize;
                    replyBuf[3] = (msgparm_t)data;
                } else {
                    replyBuf[2] = 0;
                }

                fSock.sendMsg(replyBuf, reply);
                FreeMessage(replyBuf, reply);
            }
            break;
        case kCli2Auth_FileDownloadRequest:
            {
                dsString filename = (dsChar*)(reqBuf[1]);
                const dsNetMsg* reply = GET_Auth2Cli(kAuth2Cli_FileDownloadChunk);
                replyBuf = AllocMessage(reply);
                replyBuf[0] = reqBuf[0];    // Trans ID

                // Get the file data and send it in chunks to the client
                dsAuthFileData fileData;
                fileData.read(filename.replace("\\", "/"));
                replyBuf[1] = fileData.getStatus();
                dsUByte* data = NULL;
                dsLogF(kLogInfo, "Sending file %s", filename.cstr().data());
                while (fileData.getPos() < fileData.getTotalSize()) {
                    if (fileData.getStatus() == kNetSuccess) {
                        replyBuf[2] = fileData.getTotalSize();
                        replyBuf[3] = fileData.getPos();
                        size_t chunkSize = replyBuf[2] - replyBuf[3];
                        if (chunkSize > 16384)  // Send 16KB at a time
                            chunkSize = 16384;
                        replyBuf[4] = chunkSize;
                        data = new dsUByte[chunkSize];
                        fileData.write(data, chunkSize);
                        replyBuf[1] = fileData.getStatus();  // In case the write() fails
                        replyBuf[5] = (msgparm_t)data;
                    } else {
                        replyBuf[2] = 0;
                        replyBuf[3] = 0;
                        replyBuf[4] = 0;
                        data = NULL;
                        fileData.setFinished(); // Stop sending chunks if we get an error
                    }

                    if (!fSock.sendMsg(replyBuf, reply)) {
                        dsLogF(kLogWarning, "Send of %s failed.  Aborting transfer",
                                            filename.cstr().data());
                        fileData.setFinished();
                    }

                    // Do this by hand so we can re-use the message buffer
                    replyBuf[5] = 0;
                    if (data != NULL)
                        delete[] data;
                }
                FreeMessage(replyBuf, reply);
            }
            break;
        case kCli2Auth_FileDownloadChunkAck:
            // Just ignore it for now...
            break;
        //case kCli2Auth_PropagateBuffer:
        //case kCli2Auth_GetPublicAgeList:
        //case kCli2Auth_SetAgePublic:
        case kCli2Auth_LogPythonTraceback:
            {
                dsString traceback = (dsChar*)(reqBuf[0]);
                dsLogF(kLogInfo, "Client got python crash:\n----------\n%s\n----------",
                                 traceback.cstr().data());
            }
            break;
        case kCli2Auth_LogStackDump:
            {
                dsString stackdump = (dsChar*)(reqBuf[0]);
                dsLogF(kLogInfo, "Client got stackdump:\n----------\n%s\n----------",
                                 stackdump.cstr().data());
            }
            break;
        case kCli2Auth_LogClientDebuggerConnect:
            {
                dsLogF(kLogInfo, "Client Attached Debugger.  Param: %u", reqBuf[0]);
            }
            break;
        //case kCli2Auth_LogPythonTraceback:
        //case kCli2Auth_LogStackDump:
        //case kCli2Auth_LogClientDebuggerConnect:
        //case kCli2Auth_ScoreCreate:
        //case kCli2Auth_ScoreDelete:
        //case kCli2Auth_ScoreGetScores:
        //case kCli2Auth_ScoreAddPoints:
        //case kCli2Auth_ScoreTransferPoints:
        //case kCli2Auth_ScoreSetPoints:
        //case kCli2Auth_ScoreGetRanks:
        default:
            dsLogF(kLogError, "Got unsupported message type: %u", msgType);
            // Continue anyway
        }

        FreeMessage(reqBuf, msg);
    }

    fSock.flush();
    dsLog(kLogInfo, "Auth Server closed");
}

bool dsAuthServer::calculateKeys(const dsBigInteger& Y)
{
    dsString privateKeyPath = dsConfigMgr::GetPath("AuthPrivateKey");
    dsString publicKeyPath = dsConfigMgr::GetPath("AuthPublicKey");

    // Server's xor key for client
    dsUByte tempKey[] = { 0x4F, 0x17, 0xC8, 0x19, 0x3D, 0x08, 0xF3 };
    memcpy(fServerSeed, tempKey, 7);

    // Server communication encryption
    dsUByte k_seed[64];
    FILE* keyF = fopen(privateKeyPath.cstr(), "rb");
    if (keyF == NULL)
        return false;
    fread(k_seed, sizeof(dsUByte), 64, keyF);
    fclose(keyF);

    dsUByte n_seed[64];
    keyF = fopen(publicKeyPath.cstr(), "rb");
    if (keyF == NULL)
        return false;
    fread(n_seed, sizeof(dsUByte), 64, keyF);
    fclose(keyF);

    dsBigInteger K(k_seed, 64);
    dsBigInteger N(n_seed, 64);
    dsBigInteger clientSeed = Y.PowMod(K, N);

    dsUByte* seedData;
    size_t seedSize;
    clientSeed.getData(seedData, seedSize);
    if (seedSize > 7)
        seedSize = 7;
    for (size_t i=0; i<seedSize; i++)
        fSharedKey[i] = seedData[i] ^ fServerSeed[i];
    for (size_t i=seedSize; i<7; i++)
        fSharedKey[i] = fServerSeed[i];
    delete[] seedData;

    // "Droid" Key for Python and SDL files
    s_droidKey[0] = dsConfigMgr::GetUInt("DataKey1");
    s_droidKey[1] = dsConfigMgr::GetUInt("DataKey2");
    s_droidKey[2] = dsConfigMgr::GetUInt("DataKey3");
    s_droidKey[3] = dsConfigMgr::GetUInt("DataKey4");
    return true;
}

void dsAuthServer::login(const char* username, const dsSha1Hash& passhash)
{
    fLoginInfo.fAcctFlags = 0;
    fLoginInfo.fBillingType = 0;

    if (!fAuthDb.isConnected()) {
        fLoginInfo.fResult = kNetErrInternalError;
        fLoginInfo.fPrivLevel = kPrivNormal;
        return;
    }

    char safeUser[64];
    fAuthDb.escapeString(safeUser, username, 64);
    dsPostgresResult res = fAuthDb.query(dsString::Format(
        "SELECT * FROM auth.\"Accounts\" WHERE LOWER(\"Login\")=LOWER('%s')", safeUser));
    if (!res.isValid()) {
        dsLogF(kLogError, "SQL Error in login: %s", fAuthDb.getErrorMsg());
        fLoginInfo.fResult = kNetErrInternalError;
        fLoginInfo.fPrivLevel = kPrivNormal;
        return;
    }

    if (res.getNumRows() == 0) {
        fLoginInfo.fResult = kNetErrAccountNotFound;
        fLoginInfo.fPrivLevel = kPrivNormal;
        return;
    } else if (res.getNumRows() > 1) {
        fLoginInfo.fResult = kNetErrInternalError;
        fLoginInfo.fPrivLevel = kPrivNormal;
        return;
    }

    dsPostgresRow row = res.fetchRow();
    fLoginInfo.fAcctUuid = dsUuid::fromString(row["\"AcctUuid\""].getString());
    fLoginInfo.fAcctFlags = row["\"AcctFlags\""].getUInt();
    fLoginInfo.fBillingType = row["\"BillingType\""].getUInt();
    fLoginInfo.fPrivLevel = (EPrivilegeLevel)row["\"PrivLevel\""].getInt();
    if (passhash.toString() != row["\"PassHash\""].getString()) {
        fLoginInfo.fResult = kNetErrAuthenticationFailed;
        fLoginInfo.fPrivLevel = kPrivNormal;
        return;
    }
    if (fLoginInfo.fPrivLevel == kPrivBanned)
        fLoginInfo.fResult = kNetErrAccountBanned;
    else
        fLoginInfo.fResult = kNetSuccess;
}

#include "dsAuthServer_VaultComm.cpp"
