#include "dsRC4.h"
#include "../dsLog.h"
#include <algorithm>

void dsRC4Socket::init(size_t keySize, dsUByte* keyData)
{
    if (keyData != NULL) {
        size_t x = 0;
        for (size_t i=0; i<256; i++) {
            x = (keyData[i % keySize] + fSend.fState[i] + x) & 0xFF;
            std::swap(fSend.fState[i], fSend.fState[x]);
        }

        x = 0;
        for (size_t i=0; i<256; i++) {
            x = (keyData[i % keySize] + fRecv.fState[i] + x) & 0xFF;
            std::swap(fRecv.fState[i], fRecv.fState[x]);
        }
        fEncrypted = true;
    } else {
        fEncrypted = false;
    }
}

void dsRC4Socket::_crypt::crypt(size_t size, const dsUByte* src, dsUByte* dest)
{
    for (size_t i=0; i<size; i++) {
        fX = (fX + 1) & 0xFF;
        fY = (fState[fX] + fY) & 0xFF;
        std::swap(fState[fX], fState[fY]);
        dest[i] = src[i] ^ fState[(fState[fX] + fState[fY]) & 0xFF];
    }
}

ssize_t dsRC4Socket::send(const void* buf, size_t size)
{
/*
#ifdef COMMDEBUG
    dsString ln = "<SEND> ";
    char* sBuf = new char[(3 * size) + 1];
    for (size_t i=0; i<size; i++)
        sprintf(sBuf + (3*i), "%02X ", ((dsUByte*)buf)[i]);
    sBuf[(3 * size)] = 0;
    ln += sBuf;
    delete[] sBuf;
    dsLog(kLogDebug, ln.cstr());
#endif
*/

    ssize_t sSize;
    if (!fEncrypted) {
        sSize = dsAsyncSocket::send(buf, size);
    } else {
        dsUByte* cBuf = new dsUByte[size];
        lockSend();
        fSend.crypt(size, (const dsUByte*)buf, cBuf);
        sSize = dsAsyncSocket::send(cBuf, size);
        unlockSend();
        delete[] cBuf;
    }
    return sSize;
}

ssize_t dsRC4Socket::recv(void* buf, size_t size)
{
    ssize_t rSize;
    if (!fEncrypted) {
        rSize = dsAsyncSocket::recv(buf, size);
    } else {
        lockRecv();
        rSize = dsAsyncSocket::recv(buf, size);
        if (rSize > 0) {
            dsUByte* cBuf = new dsUByte[rSize];
            fRecv.crypt(rSize, (const dsUByte*)buf, cBuf);
            memcpy(buf, cBuf, rSize);
            delete[] cBuf;
        }
        unlockRecv();
    }

/*
#ifdef COMMDEBUG
    if (rSize > 0) {
        dsString ln = ">RECV< ";
        char* sBuf = new char[(3 * rSize) + 1];
        for (ssize_t i=0; i<rSize; i++)
            sprintf(sBuf + (3*i), "%02X ", ((dsUByte*)buf)[i]);
        sBuf[(3 * rSize)] = 0;
        ln += sBuf;
        delete[] sBuf;
        dsLog(kLogDebug, ln.cstr());
    }
#endif
*/
    return rSize;
}

ssize_t dsRC4Socket::peek(void* buf, size_t size)
{
    if (!fEncrypted)
        return dsAsyncSocket::peek(buf, size);

    dsLog(kLogError, "Cannot peek on an encrypted socket");
    return -1;
}

dsString dsRC4Socket::recvString(size_t maxlen)
{
    dsChar* buf = new dsChar[maxlen + 1];
    buf[maxlen] = 0;

    dsUInt16 size;
    if (recv(&size, sizeof(dsUInt16)) <= 0)
        size = 0;

    dsString str;
    if (size > 0) {
        recv(buf, (size > maxlen ? maxlen : size) * sizeof(dsChar));
        str.setFrom(buf, size);
        str.compact();
    }
    if (size > maxlen) {
        dsChar* cut = new dsChar[size - maxlen];
        recv(cut, (size - maxlen) * sizeof(dsChar));
        delete[] cut;
    }

    delete[] buf;
    return str;
}

static void RecvBasic(dsRC4Socket* sock, msgparm_t& data,
                      dsUInt32 size, dsUInt32 count)
{
    // Also works for floats and doubles

    if (count == 0) {
        if (size == 1) {
            dsUByte v;
            sock->recv(&v, sizeof(dsUByte));
            data = v;
        } else if (size == 2) {
            dsUInt16 v;
            sock->recv(&v, sizeof(dsUInt16));
            data = v;
        } else if (size == 4) {
            dsUInt32 v;
            sock->recv(&v, sizeof(dsUInt32));
            data = v;
        } else if (size == 8) {
            dsUInt64 v;
            sock->recv(&v, sizeof(dsUInt64));
            data = v;
        }

#ifdef COMMDEBUG
        dsLogF(kLogDebug, "     <- Int%d: %lu", size * 8, data);
#endif
    } else {
        if (size == 1)
            sock->recv((dsUByte*)data, count * sizeof(dsUByte));
        else if (size == 2)
            sock->recv((dsUByte*)data, count * sizeof(dsUInt16));
        else if (size == 4)
            sock->recv((dsUByte*)data, count * sizeof(dsUInt32));
        else if (size == 8)
            sock->recv((dsUByte*)data, count * sizeof(dsUInt64));

#ifdef COMMDEBUG
        dsString ln = dsString::Format("     <- Int%d[%d]: ", size * 8, count);
        size_t lnbufSize = count * ((size * 2) + 1);
        char* lnbuf = new char[lnbufSize + 1];
        for (size_t i=0; i<count; i++) {
            if (size == 1)
                sprintf(lnbuf + (i * 3), "%02X ", ((dsUByte*)data)[i]);
            else if (size == 2)
                sprintf(lnbuf + (i * 5), "%04X ", ((dsUInt16*)data)[i]);
            else if (size == 4)
                sprintf(lnbuf + (i * 9), "%08X ", ((dsUInt32*)data)[i]);
            else if (size == 8)
                sprintf(lnbuf + (i * 17), "%016llX ",((dsUInt64*)data)[i]);
        }
        lnbuf[lnbufSize] = 0;
        ln += lnbuf;
        delete[] lnbuf;
        dsLog(kLogDebug, ln.cstr());
#endif
    }
}

static void SendBasic(dsUByte*& buf, const msgparm_t& data,
                      dsUInt32 size, dsUInt32 count)
{
    // Also works for floats and doubles

    if (count == 0) {
        if (size == 1) {
            *(dsUByte*)buf = data;
            buf += sizeof(dsUByte);
        } else if (size == 2) {
            *(dsUInt16*)buf = data;
            buf += sizeof(dsUInt16);
        } else if (size == 4) {
            *(dsUInt32*)buf = data;
            buf += sizeof(dsUInt32);
        } else if (size == 8) {
            *(dsUInt64*)buf = data;
            buf += sizeof(dsUInt64);
        }

#ifdef COMMDEBUG
        dsLogF(kLogDebug, "     -> Int%d: %lu", size * 8, data);
#endif
    } else {
        if (size == 1) {
            memcpy(buf, (const dsUByte*)data, count * sizeof(dsUByte));
            buf += count * sizeof(dsUByte);
        } else if (size == 2) {
            memcpy(buf, (const dsUByte*)data, count * sizeof(dsUInt16));
            buf += count * sizeof(dsUInt16);
        } else if (size == 4) {
            memcpy(buf, (const dsUByte*)data, count * sizeof(dsUInt32));
            buf += count * sizeof(dsUInt32);
        } else if (size == 8) {
            memcpy(buf, (const dsUByte*)data, count * sizeof(dsUInt64));
            buf += count * sizeof(dsUInt64);
        }

#ifdef COMMDEBUG
        dsString ln = dsString::Format("     <- Int%d[%d]: ", size * 8, count);
        size_t lnbufSize = count * ((size * 2) + 1);
        char* lnbuf = new char[lnbufSize + 1];
        for (size_t i=0; i<count; i++) {
            if (size == 1)
                sprintf(lnbuf + (i * 3), "%02X ", ((dsUByte*)data)[i]);
            else if (size == 2)
                sprintf(lnbuf + (i * 5), "%04X ", ((dsUInt16*)data)[i]);
            else if (size == 4)
                sprintf(lnbuf + (i * 9), "%08X ", ((dsUInt32*)data)[i]);
            else if (size == 8)
                sprintf(lnbuf + (i * 17), "%016llX ", ((dsUInt64*)data)[i]);
        }
        lnbuf[lnbufSize] = 0;
        ln += lnbuf;
        delete[] lnbuf;
        dsLog(kLogDebug, ln.cstr());
#endif
    }
}

bool dsRC4Socket::sendMsg(const msgparm_t* data, const dsNetMsg* msg)
{
    if (msg == NULL)
        return false;

#ifdef COMMDEBUG
    dsLogF(kLogDebug, "<SEND> %s", msg->fMsgName);
#endif

    size_t bufSize = sizeof(dsUInt16) + MessageSize(data, msg);
    dsUByte* buffer = new dsUByte[bufSize];
    dsUByte* bp = buffer;
    *(dsUInt16*)bp = msg->fMsgId;
    bp += sizeof(dsUInt16);

    dsUInt32 size = 0;
    dsUInt32 count = 0;
    for (size_t i=0; i<msg->fFieldCount; i++) {
        const dsNetMsgField* field = &msg->fFields[i];
        switch (field->fType) {
        case kFieldInteger:
        case kFieldFloat:
            SendBasic(bp, data[i], field->fSize, field->fCount);
            break;
        case kFieldString:
            {
                dsUInt16 len = dsStrLen((const dsChar*)(data[i]));
                *(dsUInt16*)bp = len;
                bp += sizeof(dsUInt16);
                memcpy(bp, (const dsUByte*)(data[i]), len * sizeof(dsChar));
                bp += len * sizeof(dsChar);
#ifdef COMMDEBUG
                dsLogF(kLogDebug, "     -> Str: %s",
                       dsString((const dsChar*)(data[i]), len).cstr().data());
#endif
            }
            break;
        case kFieldVarCount:
            size = field->fSize;
            count = data[i];
            *(dsUInt32*)bp = count;
            bp += sizeof(dsUInt32);
#ifdef COMMDEBUG
            dsLogF(kLogDebug, "     -> Count: %d", count);
#endif
            break;
        case kFieldVarPtr:
        case kFieldRawVarPtr:
            memcpy(bp, (const dsUByte*)(data[i]), size * count);
            bp += size * count;
#ifdef COMMDEBUG
            {
                dsString ln = "     -> VarData: ";
                char* sBuf = new char[(3 * (size * count)) + 1];
                for (size_t j=0; j<(size * count); j++)
                    sprintf(sBuf + (3*j), "%02X ", ((dsUByte*)(data[i]))[j]);
                sBuf[3 * (size * count)] = 0;
                ln += sBuf;
                delete[] sBuf;
                dsLog(kLogDebug, ln.cstr());
            }
#endif
            size = 0;
            count = 0;
            break;
        case kFieldData:
        case kFieldPtr:
        case kFieldRawData:
        case kFieldRawPtr:
            memcpy(bp, (const dsUByte*)(data[i]), field->fSize * field->fCount);
            bp += field->fSize * field->fCount;
#ifdef COMMDEBUG
            {
                dsString ln = "     -> Data: ";
                char* sBuf = new char[(3 * (field->fSize * field->fCount)) + 1];
                for (size_t j=0; j<(field->fSize * field->fCount); j++)
                    sprintf(sBuf + (3*j), "%02X ", ((dsUByte*)(data[i]))[j]);
                sBuf[3 * (field->fSize * field->fCount)] = 0;
                ln += sBuf;
                delete[] sBuf;
                dsLog(kLogDebug, ln.cstr());
            }
#endif
            break;
        }
    }
    bool result = (send(buffer, bufSize) > 0);
    delete[] buffer;
    return result;
}

msgparm_t* dsRC4Socket::recvMsg(const dsNetMsg* msg)
{
    if (msg == NULL)
        return NULL;

#ifdef COMMDEBUG
    dsLogF(kLogDebug, "<RECV> %s", msg->fMsgName);
#endif
    msgparm_t* data = AllocMessage(msg);

    dsUInt32 size = 0;
    dsUInt32 count = 0;
    for (size_t i=0; i<msg->fFieldCount; i++) {
        const dsNetMsgField* field = &msg->fFields[i];
        switch (field->fType) {
        case kFieldInteger:
        case kFieldFloat:
            RecvBasic(this, data[i], field->fSize, field->fCount);
            break;
        case kFieldString:
            {
                dsUInt16 len;
                recv(&len, sizeof(dsUInt16));
                dsChar* str = new dsChar[len + 1];
                recv(str, len * sizeof(dsChar));
                str[len] = 0;
                data[i] = (msgparm_t)str;
#ifdef COMMDEBUG
                dsLogF(kLogDebug, "     <- Str: %s",
                       dsString((const dsChar*)(data[i]), len).cstr().data());
#endif
            }
            break;
        case kFieldVarCount:
            size = field->fSize;
            recv(&count, sizeof(dsUInt32));
            data[i] = count;
#ifdef COMMDEBUG
            dsLogF(kLogDebug, "     <- Count: %d", count);
#endif
            break;
        case kFieldVarPtr:
        case kFieldRawVarPtr:
            if (data[i] != 0)
                delete[] (dsUByte*)data[i];
            data[i] = (msgparm_t)(new dsUByte[size * count]);
            recv((dsUByte*)(data[i]), size * count);
#ifdef COMMDEBUG
            {
                dsString ln = "     <- VarData: ";
                char* sBuf = new char[(3 * (size * count)) + 1];
                for (size_t j=0; j<(size * count); j++)
                    sprintf(sBuf + (3*j), "%02X ", ((dsUByte*)(data[i]))[j]);
                sBuf[3 * (size * count)] = 0;
                ln += sBuf;
                delete[] sBuf;
                dsLog(kLogDebug, ln.cstr());
            }
#endif
            size = 0;
            count = 0;
            break;
        case kFieldData:
        case kFieldPtr:
        case kFieldRawData:
        case kFieldRawPtr:
            recv((dsUByte*)(data[i]), field->fSize * field->fCount);
#ifdef COMMDEBUG
            {
                dsString ln = "     <- Data: ";
                char* sBuf = new char[(3 * (field->fSize * field->fCount)) + 1];
                for (size_t j=0; j<(field->fSize * field->fCount); j++)
                    sprintf(sBuf + (3*j), "%02X ", ((dsUByte*)(data[i]))[j]);
                sBuf[3 * (field->fSize * field->fCount)] = 0;
                ln += sBuf;
                delete[] sBuf;
                dsLog(kLogDebug, ln.cstr());
            }
#endif
            break;
        }
    }
    return data;
}

void dsRC4Socket::test()
{
    dsRC4Socket sock(NULL);
    dsUByte key[7] = { 0x12, 0x34, 0x56, 0x78, 0x9A, 0xBC, 0xDE };
    sock.init(7, key);

    dsUByte data1[16] = { 0x3F, 0x15, 0x99, 0x4A, 0x91, 0x22, 0x04, 0xCD,
                          0x23, 0xAD, 0x05, 0x3C, 0x1D, 0xEF, 0xA9, 0x69 };
    dsUByte data2[16] = { 0x3F, 0x15, 0x99, 0x4A, 0x91, 0x22, 0x04, 0xCD,
                          0x23, 0xAD, 0x05, 0x3C, 0x1D, 0xEF, 0xA9, 0x69 };
    dsUByte cdata1[16], cdata2[16];

    sock.fSend.crypt(16, data1, cdata1);
    sock.fRecv.crypt(16, data2, cdata2);

    for (size_t i=0; i<16; i++)
        printf("%02X ", cdata1[i]);
    printf("\n");
    for (size_t i=0; i<16; i++)
        printf("%02X ", cdata2[i]);
    printf("\n");
}
