#include "dsPostgres.h"
#include <cstring>

/* dsBlob */
bool dsBlob::operator==(const dsBlob& other) const
{
    if (fBlob == other.fBlob)
        return true;
    if (fBlob == NULL || other.fBlob == NULL)
        return false;
    if (fBlob->fSize != other.fBlob->fSize)
        return false;
    return memcmp(fBlob->fData, other.fBlob->fData, fBlob->fSize) == 0;
}


/* dsSQLResultRef */
void dsSQLResultRef::ref()
{
    if (this != NULL)
        fDataRefs++;
}

void dsSQLResultRef::unref()
{
    if (this != NULL) {
        fDataRefs--;
        if (fDataRefs == 0) {
            PQclear(fData);
            delete this;
        }
    }
}

/* dsPostgresRow */
dsPostgresRow::dsPostgresRow(int rowNum, dsSQLResultRef* data)
             : fRowNum(rowNum), fData(data)
{
    fData->ref();
    fNumFields = (fData != NULL) ? PQnfields(fData->fData) : 0;
}

dsPostgresValue dsPostgresRow::operator[](int idx) const
{
    if (fData != NULL && idx < fNumFields) {
        if (PQgetisnull(fData->fData, fRowNum, idx))
            return dsPostgresValue(NULL, 0);
        else
            return dsPostgresValue(PQgetvalue(fData->fData, fRowNum, idx),
                                   PQgetlength(fData->fData, fRowNum, idx));
    } else {
        return dsPostgresValue(NULL, 0);
    }
}

dsPostgresValue dsPostgresRow::operator[](const char* idx) const
{
    int col = PQfnumber(fData->fData, idx);
    if (col >= 0) {
        if (PQgetisnull(fData->fData, fRowNum, col))
            return dsPostgresValue(NULL, 0);
        else
            return dsPostgresValue(PQgetvalue(fData->fData, fRowNum, col),
                                   PQgetlength(fData->fData, fRowNum, col));
    } else {
        return dsPostgresValue(NULL, 0);
    }
}


/* dsPostgresResult */
dsPostgresResult::dsPostgresResult(PGresult* res)
                : fCurrentRow(0)
{
    fResult = new dsSQLResultRef;
    fResult->fData = res;
    fResult->fDataRefs = 1;
}

dsPostgresResult& dsPostgresResult::operator=(PGresult* res)
{
    fResult->unref();
    fResult = new dsSQLResultRef;
    fResult->fData = res;
    fResult->fDataRefs = 1;
    return *this;
}

dsPostgresRow dsPostgresResult::operator[](int idx)
{
    if (fResult == NULL || (int)idx >= PQntuples(fResult->fData))
        return dsPostgresRow(0, NULL);
    return dsPostgresRow(idx, fResult);
}

dsPostgresRow dsPostgresResult::fetchRow()
{
    if (fResult == NULL || fCurrentRow >= PQntuples(fResult->fData))
        return dsPostgresRow(0, NULL);
    return dsPostgresRow(fCurrentRow++, fResult);
}


/* dsPostgres */
bool dsPostgres::connect(const char* host, const char* port,
                         const char* username, const char* password,
                         const char* sslmode, const char* database)
{
    close();
    dsString conn = dsString::Format("host=%s port=%s dbname=%s user=%s password=%s sslmode=%s",
                                     host, port, database, username, password, sslmode);
    fConnection = PQconnectdb(conn.cstr());
    return PQstatus(fConnection) == CONNECTION_OK;
}

void dsPostgres::close()
{
    if (fConnection != NULL)
        PQfinish(fConnection);
    fConnection = NULL;
}

dsPostgresResult dsPostgres::query(const dsString& query)
{
    verifyConnection();
    return dsPostgresResult(PQexec(fConnection, query.cstr()));
}

dsUuid dsPostgres::MakeUuid()
{
    dsPostgresResult res = query("SELECT uuid_generate_v4();");
    dsPostgresRow row = res.fetchRow();
    return dsUuid::fromString(row[(int)0].getString());
}

static void b64_block(const dsUByte* in, char* out, size_t len)
{
    static const char b64_table[] =
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

    out[0] = b64_table[in[0] >> 2];
    out[1] = b64_table[((in[0] & 0x03) << 4) | ((in[1] & 0xf0) >> 4)];
    out[2] = (len > 1) ? b64_table[((in[1] & 0x0f) << 2) | ((in[2] & 0xc0) >> 6)] : '=';
    out[3] = (len > 2) ? b64_table[in[2] & 0x3f] : '=';
}

dsString Base64(const dsUByte* src, size_t len)
{
    if (src == NULL || len == 0)
        return dsString();

    dsString result;
    char blk[5];
    blk[4] = 0;
    size_t i = 0;
    while (len > 0) {
        dsUByte in[3];
        memset(in, 0, 3);
        memcpy(in, src+i, (len > 3) ? 3 : len);
        b64_block(in, blk, len);
        result += blk;

        if (len > 3)
            len -= 3;
        else
            len = 0;
        i += 3;
    }
    return result;
}

static void ub64_block(const char* in, dsUByte* out)
{
    char vin[4];
    for (size_t i=0; i<4; i++) {
        if (in[i] >= 'A' && in[i] <= 'Z')
            vin[i] = in[i] - 'A';
        else if (in[i] >= 'a' && in[i] <= 'z')
            vin[i] = (in[i] - 'a') + 26;
        else if (in[i] >= '0' && in[i] <= '9')
            vin[i] = (in[i] - '0') + 52;
        else if (in[i] == '+')
            vin[i] = 62;
        else if (in[i] == '/')
            vin[i] = 63;
        else
            vin[i] = 0;
    }

    out[0] = (vin[0] << 2) | ((vin[1] >> 4) & 0x03);
    out[1] = ((vin[1] << 4) & 0xf0) | ((vin[2] >> 2) & 0x0f);
    out[2] = ((vin[2] << 6) & 0xc0) | (vin[3] & 0x3f);
}

void UnBase64(const dsString& src, dsUByte*& dest, size_t& len)
{
    if (src.length() == 0) {
        len = 0;
        dest = NULL;
        return;
    }

    len = (src.length() * 3) / 4
        - (src[src.length()-1] == '=' ? 1 : 0)
        - (src[src.length()-2] == '=' ? 1 : 0);
    dest = new dsUByte[len];

    size_t spos = 0, dpos = 0;
    while (dpos < len) {
        dsUByte blk[3];
        ub64_block(src.cstr()+spos, blk);
        memcpy(dest+dpos, blk, ((len - dpos) > 3) ? 3 : (len - dpos));
        spos += 4;
        dpos += 3;
    }
}
