#include "dsUuid.h"
#include "dsConfigMgr.h"
#include "dsLog.h"
#include "dsSocket.h"
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <execinfo.h>

#include "Connection.h"
#include "file/dsFileServer.h"
#include "auth/dsAuthServer.h"
#include "game/dsGameServer.h"
#include "vault/dsVaultServer.h"

enum { kServerAuth, kServerFile, kServerGame, kServerVault, kNumServers };
static bool sServerUse[kNumServers] = { false, false, false, false };
static bool doShutdown = false;

/* Keep track of children so we can kill them if necessary */
std::list<pid_t> s_children;

/* Global listener socket */
dsSocket s_listenSock;

void sigbyebye(int)
{
    dsLog(kLogInfo, "Forcing shutdown...");
    //dsVaultServer::SendConnection(NULL);
    for (std::list<pid_t>::iterator i=s_children.begin(); i!=s_children.end(); i++)
        kill(*i, SIGKILL);
    doShutdown = true;
    s_listenSock.close();
}

// A shotgun: for killing zombies :)
void sigshotgun(int)
{
    signal(SIGCHLD, sigshotgun); // Reset signal listener

    int status;
    pid_t pid = wait(&status);
    if (status != 0)
        dsLogF(kLogWarning, "Child %d exited with status %d", pid, status);

    std::list<pid_t>::iterator i = s_children.begin();
    while (i != s_children.end()) {
        if (*i == pid)
            i = s_children.erase(i);
        else
            i++;
    }
}

void sigcrash(int)
{
    dsLogF(kLogError, "Server process %d segfaulted:", getpid());

    void* btarray[20];
    int btsize = backtrace(btarray, 20);
    if (dsLogProcessor::getLog() != NULL)
        backtrace_symbols_fd(btarray, btsize, dsLogProcessor::getLog()->getFD());
    else
        backtrace_symbols_fd(btarray, btsize, 2);   // stderr

    abort();
}

int main(int argc, char** argv)
{
    dsString configPath(CONFDIR);    // Default
    if (configPath[configPath.length()-1] != '/')
        configPath += "/";
    configPath += "dirtsand.conf";

    for (int i=1; i<argc; i++) {
        if (strcmp(argv[i], "--help") == 0) {
            printf("Usage: %s [options] servers\n\n", argv[0]);
            printf("Options:\n");
            printf("    --config=path  Set the location of the config file\n");
            printf("    --help         Display this help\n\n");
            printf("Servers:\n");
            printf("    auth\n");
            printf("    file\n");
            printf("    game\n");
            printf("    vault\n");
            return 0;
        } else if (strncmp(argv[i], "--config", 8) == 0) {
            dsString s(argv[i]);
            std::vector<dsString> sp = s.split('=', 1);
            if (sp.size() == 1) {
                if (++i >= argc) {
                    fprintf(stderr, "Error: No path specified\n");
                    return 1;
                }
                configPath = argv[i];
            } else {
                configPath = sp[1];
            }
        } else if (strcmp(argv[i], "auth") == 0) {
            sServerUse[kServerAuth] = true;
        } else if (strcmp(argv[i], "file") == 0) {
            sServerUse[kServerFile] = true;
        } else if (strcmp(argv[i], "game") == 0) {
            sServerUse[kServerGame] = true;
        } else if (strcmp(argv[i], "vault") == 0) {
            sServerUse[kServerVault] = true;
        } else {
            fprintf(stderr, "Unrecognized parameter: %s\n", argv[i]);
        }
    }

    if (!sServerUse[kServerAuth] && !sServerUse[kServerFile] &&
        !sServerUse[kServerVault] && !sServerUse[kServerGame]) {
        fprintf(stderr, "You must specify at least one server to start\n");
        fprintf(stderr, "For details, see --help\n\n");
        return 1;
    }
    if (!dsConfigMgr::LoadConfig(configPath.cstr()))
        return 1;

    dsString logpath = dsConfigMgr::GetPath("LogDir", "ListenServer");
    dsLogProcessor* listenLog = dsLogProcessor::Create(logpath.cstr(),
                                (ELogType)dsConfigMgr::GetUInt("LogLevel", LOGLEVEL_DEFAULT));
    dsLogProcessor::setLog(listenLog);
    dsLog(kLogInfo, "Listen server started");

    // Handle various signals gracefully ;)
    signal(SIGTERM, sigbyebye);
    signal(SIGINT, sigbyebye);
    signal(SIGCHLD, sigshotgun);
    signal(SIGSEGV, sigcrash);

    if (!s_listenSock.bind(dsConfigMgr::GetUInt("ListenPort", LISTENPORT_DEFAULT))) {
        fprintf(stderr, "Fatal error starting the master listen server\n");
        return 1;
    }

    // Start the Vault server if it's requested
    if (sServerUse[kServerVault]) {
        pid_t child = dsVaultServer::Create();
        s_children.push_back(child);
    }

    while (!doShutdown) {
        // Listen for connections 'n stuff
        dsSocket* conn = s_listenSock.listen(10);
        if (conn == NULL)
            continue;

        ConnectHeader hdr;
        if (conn->recv(&hdr, sizeof(hdr)) != sizeof(hdr)) {
            dsLogF(kLogError, "Invalid Connection Header packet from %s",
                              conn->getRemoteIpStr().cstr().data());
            continue;
        }
        switch (hdr.fConnType) {
        case kConnTypeCliToAuth:
            if (sServerUse[kServerAuth]) {
                pid_t child = dsAuthServer::Create(conn);
                s_children.push_back(child);
            } else {
                dsLog(kLogError, "Got an Auth connection, but this server is not hosting Auth");
                delete conn;
            }
            break;
        case kConnTypeCliToFile:
            if (sServerUse[kServerFile]) {
                pid_t child = dsFileServer::Create(conn);
                s_children.push_back(child);
            } else {
                dsLog(kLogError, "Got a File connection, but this server is not hosting files");
                delete conn;
            }
            break;
        case kConnTypeCliToGame:
            if (sServerUse[kServerGame]) {
                dsGameServer::AcquireClient(conn);
            } else {
                dsLog(kLogError, "Got a Game connection, but this server is not hosting Games");
                delete conn;
            }
            break;
        case kConnTypeVault_DS:
            if (sServerUse[kServerVault]) {
                dsVaultServer::SendConnection(conn);
            } else {
                dsLog(kLogError, "Got a Vault connection, but this server is not hosting the Vault");
                delete conn;
            }
            break;
        default:
            dsLogF(kLogInfo, "Got unknown connection type %d from %s",
                             hdr.fConnType, conn->getRemoteIpStr().cstr().data());
            delete conn;
            break;
        }
    }

    dsLogProcessor::setLog(NULL, true);
    exit(0);
    return 0;   // Make the compiler happy
}
