#include "dsFileServer.h"
#include "../dsConfigMgr.h"
#include "../dsLog.h"
#include <unistd.h>
#include <cstdlib>
#include <errno.h>

#include "FileMessages.h"
#include "dsManifest.h"
#include "dsFileData.h"

static dsUInt32 getNextReader()
{
    static dsUInt32 s_reader = 0;
    return ++s_reader;
}

pid_t dsFileServer::Create(dsSocket* sock)
{
    pid_t pid = fork();
    if (pid < 0) {
        dsLogF(kLogError, "Error spawning File server: %s", strerror(errno));
        delete sock;
        exit(1);
    } else if (pid == 0) {
        s_listenSock.close();
        s_children.clear();
        dsString logpath = dsConfigMgr::GetPath("LogDir", "FileServer");
        dsLogProcessor* fileLog = dsLogProcessor::Create(logpath.cstr(),
                                  (ELogType)dsConfigMgr::GetUInt("LogLevel", LOGLEVEL_DEFAULT));
        fileLog->setAddress(sock->getRemoteIpStr().cstr());
        dsLogProcessor::setLog(fileLog, true);

        dsFileServer server(sock);
        server.run();
        exit(0);
    } else {
        // Get rid of this socket and go back to listening
        delete sock;
        return pid;
    }
}

void dsFileServer::run()
{
    dsLog(kLogInfo, "File server started");

    FileConnectHeader connectInfo;
    fSock.peek(&connectInfo.fHeaderSize, sizeof(dsUInt32));
    if (connectInfo.fHeaderSize != sizeof(FileConnectHeader)) {
        dsLog(kLogError, "Invalid File Server connection packet");
        return;
    }
    if (fSock.recv(&connectInfo, sizeof(connectInfo)) != sizeof(FileConnectHeader)) {
        dsLog(kLogError, "Invalid File Server connection packet");
        return;
    }
    fBuildId = connectInfo.fBuildId;

    FileMsg_Header header;
    while (fSock.recv(&header, sizeof(header)) > 0) {
        FileMsg_Header replyHdr;
        switch (header.fMsgId) {
        case kCli2File_PingRequest:
            {
                Cli2File_PingRequest req;
                fSock.recv(&req, sizeof(req));

                File2Cli_PingReply reply;
                replyHdr.fMsgSize = sizeof(replyHdr) + sizeof(reply);
                replyHdr.fMsgId = kFile2Cli_PingReply;
                reply.fPingTimeMs = req.fPingTimeMs;
                dsUByte replyBuf[replyHdr.fMsgSize];
                memcpy(replyBuf, &replyHdr, sizeof(replyHdr));
                memcpy(replyBuf + sizeof(replyHdr), &reply, sizeof(reply));
                fSock.send(replyBuf, replyHdr.fMsgSize);

                dsLog(kLogDebug, "PING!");
            }
            break;
        case kCli2File_BuildIdRequest:
            {
                Cli2File_BuildIdRequest req;
                fSock.recv(&req, sizeof(req));

                File2Cli_BuildIdReply reply;
                replyHdr.fMsgSize = sizeof(replyHdr) + sizeof(reply);
                replyHdr.fMsgId = kFile2Cli_BuildIdReply;
                reply.fTransId = req.fTransId;
                reply.fResult = kNetSuccess;
                reply.fBuildId = fBuildId;
                dsUByte replyBuf[replyHdr.fMsgSize];
                memcpy(replyBuf, &replyHdr, sizeof(replyHdr));
                memcpy(replyBuf + sizeof(replyHdr), &reply, sizeof(reply));
                fSock.send(replyBuf, replyHdr.fMsgSize);
            }
            break;
        case kCli2File_ManifestRequest:
            {
                Cli2File_ManifestRequest req;
                fSock.recv(&req, sizeof(req));

                File2Cli_ManifestReply reply;
                replyHdr.fMsgSize = sizeof(replyHdr) + sizeof(reply);
                replyHdr.fMsgId = kFile2Cli_ManifestReply;
                reply.fTransId = req.fTransId;
                reply.fReaderId = getNextReader();

                // Get and build the manifest info
                dsString group(req.fGroup, 260);  // Explicit size for security
                group.compact();
                dsManifest mfs;
                mfs.read(group + ".mfs");
                reply.fResult = mfs.getStatus();
                dsChar* data = NULL;
                dsLogF(kLogDebug, "Sending Manifest %s", group.cstr().data());
                if (mfs.getStatus() == kNetSuccess) {
                    reply.fNumFiles = mfs.getNumFiles();
                    reply.fDataSize = mfs.calcTotalSize();
                    replyHdr.fMsgSize += reply.fDataSize * sizeof(dsChar);
                    data = new dsChar[reply.fDataSize];
                    mfs.write(data, reply.fDataSize);
                    reply.fResult = mfs.getStatus();  // In case something terribly goes wrong :P
                } else {
                    reply.fNumFiles = 0;
                    reply.fDataSize = 0;
                }

                dsUByte* replyBuf = new dsUByte[replyHdr.fMsgSize];
                memcpy(replyBuf, &replyHdr, sizeof(replyHdr));
                memcpy(replyBuf + sizeof(replyHdr), &reply, sizeof(reply));
                if (data != NULL)
                    memcpy(replyBuf + sizeof(replyHdr) + sizeof(reply), data, reply.fDataSize * sizeof(dsChar));
                fSock.send(replyBuf, replyHdr.fMsgSize);
                if (data != NULL)
                    delete[] data;
                delete[] replyBuf;
            }
            break;
        case kCli2File_ManifestEntryAck:
            {
                Cli2File_ManifestEntryAck ack;
                fSock.recv(&ack, sizeof(ack));
                // Just ignore it for now...
            }
            break;
        case kCli2File_FileDownloadRequest:
            {
                Cli2File_FileDownloadRequest req;
                fSock.recv(&req, sizeof(req));

                File2Cli_FileDownloadReply reply;
                size_t baseSize = sizeof(replyHdr) + sizeof(reply);
                replyHdr.fMsgId = kFile2Cli_FileDownloadReply;
                reply.fTransId = req.fTransId;
                reply.fReaderId = getNextReader();

                // Get the file data and send it in chunks to the client
                dsString filename(req.fFilename, 260);  // Explicit size for security
                filename.compact();
                dsFileData fileData;
                fileData.read(filename);
                reply.fResult = fileData.getStatus();
                dsUByte* data = NULL;
                dsLogF(kLogInfo, "Sending file %s", filename.cstr().data());
                while (fileData.getPos() < fileData.getTotalSize()) {
                    if (fileData.getStatus() == kNetSuccess) {
                        reply.fFileSize = fileData.getTotalSize();
                        reply.fDataSize = reply.fFileSize - fileData.getPos();
                        if (reply.fDataSize > 16384) // Send 16KB at a time
                            reply.fDataSize = 16384;
                        data = new dsUByte[reply.fDataSize];
                        fileData.write(data, reply.fDataSize);
                        reply.fResult = fileData.getStatus();  // In case something terribly goes wrong :P
                    } else {
                        reply.fFileSize = 0;
                        reply.fDataSize = 0;
                        data = NULL;
                        fileData.setFinished(); // Stop sending chunks if we get an error
                    }

                    replyHdr.fMsgSize = baseSize + reply.fDataSize;
                    dsUByte* replyBuf = new dsUByte[replyHdr.fMsgSize];
                    memcpy(replyBuf, &replyHdr, sizeof(replyHdr));
                    memcpy(replyBuf + sizeof(replyHdr), &reply, sizeof(reply));
                    if (data != NULL)
                        memcpy(replyBuf + sizeof(replyHdr) + sizeof(reply), data, reply.fDataSize);
                    if (fSock.send(replyBuf, replyHdr.fMsgSize) != (ssize_t)replyHdr.fMsgSize) {
                        dsLogF(kLogWarning, "Error received while sending %s; abort transfer",
                                            filename.cstr().data());
                        fileData.setFinished();
                    }
                    if (data != NULL)
                        delete[] data;
                    delete[] replyBuf;
                }
            }
            break;
        case kCli2File_FileDownloadChunkAck:
            {
                Cli2File_FileDownloadChunkAck ack;
                fSock.recv(&ack, sizeof(ack));
                // Just ignore it for now...
            }
            break;
        default:
            dsLogF(kLogError, "Got unsupported message type: %u", header.fMsgId);
            return;
        }
    }

    fSock.flush();
    dsLog(kLogInfo, "File Server closed");
}
