#include "dsGameServer.h"
#include "../dsConfigMgr.h"
#include "../dsLog.h"
#include "../Connection.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>

/* dsGameComm */
void dsGameComm::run()
{
    dsLog(kLogInfo, "Client connected");

    NetCli_Cli2Srv_Connect cryptConn;
    if (fSock.recv(&cryptConn.fMessage, sizeof(dsUByte)) != sizeof(dsUByte)
        || cryptConn.fMessage != kNetCliCli2SrvConnect) {
        dsLog(kLogError, "Invalid NetCli connection packet");
        return;
    }
    if (fSock.recv(&cryptConn.fLength, sizeof(dsUByte)) != sizeof(dsUByte)) {
        dsLog(kLogError, "Invalid NetCli connection packet");
        return;
    }

    size_t keyDataLen = cryptConn.fLength - (sizeof(dsUByte) * 2);
    if (keyDataLen > 64) {
        dsLog(kLogError, "Seed value too long");
        return;
    }
    memset(cryptConn.fKeyYData, 0, 64);
    fSock.recv(cryptConn.fKeyYData, keyDataLen);

    NetCli_Srv2Cli_Encrypt cryptReply;
    cryptReply.fMessage = kNetCliSrv2CliEncrypt;
    cryptReply.fLength = sizeof(cryptReply);
    if (!calculateKeys(dsBigInteger(cryptConn.fKeyYData, keyDataLen))) {
        dsLog(kLogError, "Error setting up encryption keys");
        return;
    }
    memcpy(cryptReply.fSeed, fServerSeed, 7);
    fSock.send(&cryptReply, sizeof(cryptReply));
    fSock.init(7, fSharedKey);

    dsUInt16 msgType;
    msgparm_t* replyBuf;
    while (fSock.recv(&msgType, sizeof(dsUInt16)) > 0) {
        const dsNetMsg* msg = GET_Cli2Game(msgType);
        msgparm_t* reqBuf = fSock.recvMsg(msg);

        switch (msgType) {
        default:
            dsLogF(kLogError, "Got unsupported message type: %u", msgType);
            // Continue anyway
        }

        FreeMessage(reqBuf, msg);
    }

    fSock.flush();
    dsLog(kLogInfo, "Client disconnected");
}


bool dsGameComm::calculateKeys(const dsBigInteger& Y)
{
    dsString privateKeyPath = dsConfigMgr::GetPath("GamePrivateKey");
    dsString publicKeyPath = dsConfigMgr::GetPath("GamePublicKey");

    // Server's xor key for client
    dsUByte tempKey[] = { 0x39, 0xA4, 0x2C, 0xD3, 0x91, 0x1E, 0x08 };
    memcpy(fServerSeed, tempKey, 7);

    // Server communication encryption
    dsUByte k_seed[64];
    FILE* keyF = fopen(privateKeyPath.cstr(), "rb");
    if (keyF == NULL)
        return false;
    fread(k_seed, sizeof(dsUByte), 64, keyF);
    fclose(keyF);

    dsUByte n_seed[64];
    keyF = fopen(publicKeyPath.cstr(), "rb");
    if (keyF == NULL)
        return false;
    fread(n_seed, sizeof(dsUByte), 64, keyF);
    fclose(keyF);

    dsBigInteger K(k_seed, 64);
    dsBigInteger N(n_seed, 64);
    dsBigInteger clientSeed = Y.PowMod(K, N);

    dsUByte* seedData;
    size_t seedSize;
    clientSeed.getData(seedData, seedSize);
    if (seedSize > 7)
        seedSize = 7;
    for (size_t i=0; i<seedSize; i++)
        fSharedKey[i] = seedData[i] ^ fServerSeed[i];
    for (size_t i=seedSize; i<7; i++)
        fSharedKey[i] = fServerSeed[i];
    delete[] seedData;

    return true;
}


/* dsGameServer */
std::list<dsGameServer::GameServerAccess> dsGameServer::fServers;
dsMutex dsGameServer::fServerListMutex;

int dsGameServer::Create(const dsUuid& ageId)
{
    int pipefd[2];
    if (socketpair(AF_UNIX, SOCK_STREAM, 0, pipefd) < 0) {
        dsLogF(kLogError, "Error creating Game pipe: %s", strerror(errno));
        exit(1);
    }

    pid_t pid = fork();
    if (pid < 0) {
        dsLogF(kLogError, "Error spawning Game server: %s", strerror(errno));
        close(pipefd[0]);
        close(pipefd[1]);
        exit(1);
    } else if (pid == 0) {
        s_listenSock.close();
        s_children.clear();
        dsString logpath = dsConfigMgr::GetPath("LogDir", "GameServer");
        dsLogProcessor* gameLog = dsLogProcessor::Create(logpath.cstr(),
                                  (ELogType)dsConfigMgr::GetUInt("LogLevel", LOGLEVEL_DEFAULT));
        dsLogProcessor::setLog(gameLog, true);

        dsGameServer server(pipefd[0], ageId);
        close(pipefd[1]);   // Child only uses read end
        server.run();
        exit(0);
    } else {
        GameServerAccess acc;
        acc.fConnPipe = pipefd[1];
        acc.fAgeInstance = ageId;
        close(pipefd[0]);   // Parent only uses write end

        fServerListMutex.lock();
        fServers.push_back(acc);
        fServerListMutex.unlock();
        s_children.push_back(pid);
        return acc.fConnPipe;
    }
}

void dsGameServer::AcquireClient(dsSocket* sock)
{
    GameConnectHeader connectInfo;
    sock->peek(&connectInfo.fHeaderSize, sizeof(dsUInt32));
    if (connectInfo.fHeaderSize != sizeof(connectInfo)) {
        dsLog(kLogError, "Invalid Game Server connection packet");
        delete sock;
        return;
    }
    if (sock->recv(&connectInfo, sizeof(connectInfo)) != sizeof(connectInfo)) {
        dsLog(kLogError, "Invalid Game Server connection packet");
        delete sock;
        return;
    }

    // Search for a game server, or create one
    int pipe = -1;
    for (std::list<GameServerAccess>::iterator it = fServers.begin(); it != fServers.end(); it++) {
        if (it->fAgeInstance == connectInfo.fAgeUuid) {
            pipe = it->fConnPipe;
            break;
        }
    }
    if (pipe == -1)
        pipe = Create(connectInfo.fAgeUuid);

    msghdr msg;
    iovec iov[1];
    char ccmsg[CMSG_SPACE(sizeof(int))];

    msg.msg_control = ccmsg;
    msg.msg_controllen = sizeof(ccmsg);
    msg.msg_name = NULL;
    msg.msg_namelen = 0;

    cmsghdr* cmptr = CMSG_FIRSTHDR(&msg);
    cmptr->cmsg_len = CMSG_LEN(sizeof(int));
    cmptr->cmsg_level = SOL_SOCKET;
    cmptr->cmsg_type = SCM_RIGHTS;
    *(int*)CMSG_DATA(cmptr) = (sock == NULL) ? 0 : sock->getHandle();

    // Stupidity: Needs at least one byte to send...  >.>
    iov[0].iov_base = const_cast<char*>("x");
    iov[0].iov_len = 1;
    msg.msg_iov = iov;
    msg.msg_iovlen = 1;

    if (sendmsg(pipe, &msg, 0) < 0)
        dsLogF(kLogError, "Error sending FD: %s", strerror(errno));
    if (sock != NULL)
        delete sock;
}

int dsGameServer::getClientConnection()
{
    msghdr msg;
    iovec iov[1];
    char ccmsg[CMSG_SPACE(sizeof(int))];
    char buf[1];

    msg.msg_control = ccmsg;
    msg.msg_controllen = sizeof(ccmsg);
    msg.msg_name = NULL;
    msg.msg_namelen = 0;

    // For getting the dummy byte
    iov[0].iov_base = buf;
    iov[0].iov_len = 1;
    msg.msg_iov = iov;
    msg.msg_iovlen = 1;

    if (recvmsg(fListenPipe, &msg, 0) < 0) {
        dsLogF(kLogError, "Got invalid FD message: %s", strerror(errno));
        return -1;
    }

    cmsghdr* cmptr = CMSG_FIRSTHDR(&msg);
    if (cmptr->cmsg_type != SCM_RIGHTS) {
        dsLogF(kLogError, "Got invalid FD message type: %d", cmptr->cmsg_type);
        return -1;
    }
    return *(int*)CMSG_DATA(cmptr);
}

void dsGameServer::run()
{
    dsLogF(kLogInfo, "Game server started");

    while (true) {
        int cliSock = getClientConnection();
        if (cliSock < 0)
            break;
        dsGameComm* client = new dsGameComm(cliSock, this);
        client->start();
        fClientListMutex.lock();
        fClients.push_back(client);
        fClientListMutex.unlock();
    }

    dsLog(kLogInfo, "Game Server closed");
}

void dsGameServer::destroyConnection(dsGameComm* self)
{
    fClientListMutex.lock();
    std::list<dsGameComm*>::iterator ci = fClients.begin();
    while (ci != fClients.end()) {
        if (*ci == self) {
            ci = fClients.erase(ci);
            delete *ci;
        } else {
            ci++;
        }
    }

    bool finished = fClients.empty();
    fClientListMutex.unlock();
    if (finished) {
        // Signal the server to stop
        close(fListenPipe);
    }
}
