#include "dsVaultClient.h"
#include "../dsLog.h"
#include "../dsConfigMgr.h"
#include "../Connection.h"

#include "dsVaultServer.h"

static dsUInt32 nextTransId()
{
    static dsUInt32 sTransId = 0;
    return ++sTransId;
}

dsSocket* dsVaultClient::getConnection()
{
    dsString vaultHost = dsConfigMgr::GetString("VaultHost", VAULTHOST_DEFAULT);
    dsUInt16 vaultPort = dsConfigMgr::GetUInt("ListenPort", LISTENPORT_DEFAULT);
    dsSocket* sock = new dsSocket();
    if (!sock->connect(vaultHost.cstr(), vaultPort)) {
        dsLog(kLogError, "Error connecting to Vault server");
        delete sock;
        return NULL;
    }
    return sock;
}

void dsVaultClient::run()
{
    // Set up the connection to the vault server
    ConnectHeader header;
    header.fConnType = kConnTypeVault_DS;
    header.fSockHeaderSize = sizeof(header);
    header.fBuildId = 0;
    header.fBuildType = 0;
    header.fBranchId = 0;
    fSock.send(&header, sizeof(header));

    VaultConnectHeader connectInfo;
    connectInfo.fHeaderSize = sizeof(connectInfo);
    connectInfo.fToken = dsVaultServer::getAuthToken();
    fSock.send(&connectInfo, sizeof(connectInfo));

    // Make sure everything is good so far:
    fSock.flush();
    if (!fSock.isConnected()) {
        dsLog(kLogError, "Error initializing vault client");
        fStatus = kNetErrConnectFailed;
        return;
    }

    // Set up encryption to the vault server
    dsUByte y_data[64];
    if (!calculateKeys(y_data)) {
        dsLog(kLogError, "Error setting up encryption parameters");
        fStatus = kNetErrConnectFailed;
        return;
    }

    NetCli_Cli2Srv_Connect cryptConn;
    cryptConn.fMessage = kNetCliCli2SrvConnect;
    cryptConn.fLength = sizeof(cryptConn);
    memcpy(cryptConn.fKeyYData, y_data, 64);
    fSock.send(&cryptConn, sizeof(cryptConn));

    NetCli_Srv2Cli_Encrypt cryptReply;
    if ((fSock.recv(&cryptReply, sizeof(cryptReply)) != sizeof(cryptReply))
        || (cryptReply.fMessage != kNetCliSrv2CliEncrypt)) {
        dsLog(kLogError, "Error getting encryption reply from server");
        fStatus = kNetErrConnectFailed;
        return;
    }
    startCrypt(cryptReply.fSeed);

    // TODO: Start Keepalive thread

    dsLog(kLogInfo, "Connected to vault");
    fStatus = kNetSuccess;

    dsUInt16 msgType;
    while (!fFinished && fSock.recv(&msgType, sizeof(dsUInt16)) > 0) {
        const dsNetMsg* msg = GET_Vault2Cli(msgType);
        msgparm_t* reqBuf = fSock.recvMsg(msg);

        switch (msgType) {
        case kVault2Cli_PingReply:
            // TODO: Pong!
            break;
        case kVault2Cli_VaultNodeCreated:
            fCallback->onNodeCreated(reqBuf);
            break;
        case kVault2Cli_VaultNodeFetched:
            fCallback->onNodeFetched(reqBuf);
            break;
        case kVault2Cli_VaultNodeChanged:
            fCallback->onNodeChanged(reqBuf);
            break;
        //case kVault2Cli_VaultNodeDeleted:
        case kVault2Cli_VaultNodeAdded:
            fCallback->onNodeAdded(reqBuf);
            break;
        //case kVault2Cli_VaultNodeRemoved:
        case kVault2Cli_VaultNodeRefsFetched:
            fCallback->onNodeRefsFetched(reqBuf);
            break;
        case kVault2Cli_VaultNodeFindReply:
            fCallback->onNodeFindReply(reqBuf);
            break;
        case kVault2Cli_VaultSaveNodeReply:
            fCallback->onSaveNodeReply(reqBuf);
            break;
        case kVault2Cli_VaultAddNodeReply:
            fCallback->onAddNodeReply(reqBuf);
            break;
        //case kVault2Cli_VaultRemoveNodeReply:
        default:
            dsLogF(kLogError, "Got unsupported vault message %d", msgType);
            // Continue anyway
        }

        FreeMessage(reqBuf, msg);
    }
}

bool dsVaultClient::calculateKeys(dsUByte* y_data)
{
    dsString sharedKeyPath = dsConfigMgr::GetPath("VaultSharedKey");
    dsString publicKeyPath = dsConfigMgr::GetPath("VaultPublicKey");

    // Server communication encryption
    dsUByte x_seed[64];
    FILE* keyF = fopen(sharedKeyPath.cstr(), "rb");
    if (keyF == NULL)
        return false;
    fread(x_seed, sizeof(dsUByte), 64, keyF);
    fclose(keyF);

    dsUByte n_seed[64];
    keyF = fopen(publicKeyPath.cstr(), "rb");
    if (keyF == NULL)
        return false;
    fread(n_seed, sizeof(dsUByte), 64, keyF);
    fclose(keyF);

    dsBigInteger X(x_seed, 64);
    dsBigInteger N(n_seed, 64);
    dsBigInteger b = dsBigInteger::Random(512);
    fClientSeed = X.PowMod(b, N);
    fServerSeed = dsBigInteger(4).PowMod(b, N);

    dsUByte* tempY;
    size_t sizeY;
    fServerSeed.getData(tempY, sizeY);
    memset(y_data, 0, 64);
    memcpy(y_data, tempY, (sizeY < 64) ? sizeY : 64);
    delete[] tempY;

    return true;
}

void dsVaultClient::startCrypt(dsUByte* seed)
{
    dsUByte* seedData;
    size_t seedSize;
    dsUByte sharedKey[7];
    fClientSeed.getData(seedData, seedSize);
    for (size_t i=0; i<7; i++)
        sharedKey[i] = seedData[i] ^ seed[i];
    delete[] seedData;

    fSock.init(7, sharedKey);
}

dsUInt32 dsVaultClient::sendCreateRequest(dsUByte* nodeData, dsUInt32 nodeSize)
{
    waitForStatus();

    const dsNetMsg* request = GET_Cli2Vault(kCli2Vault_VaultNodeCreate);
    msgparm_t* reqBuf = AllocMessage(request);
    dsUInt32 transId = nextTransId();
    reqBuf[0] = transId;
    reqBuf[1] = nodeSize;
    reqBuf[2] = (msgparm_t)nodeData;
    fSock.sendMsg(reqBuf, request);

    // Don't delete twice
    reqBuf[2] = 0;
    FreeMessage(reqBuf, request);
    return transId;
}

dsUInt32 dsVaultClient::sendFetchRequest(dsInt32 nodeId)
{
    waitForStatus();

    const dsNetMsg* request = GET_Cli2Vault(kCli2Vault_VaultNodeFetch);
    msgparm_t* reqBuf = AllocMessage(request);
    dsUInt32 transId = nextTransId();
    reqBuf[0] = transId;
    reqBuf[1] = nodeId;
    fSock.sendMsg(reqBuf, request);

    FreeMessage(reqBuf, request);
    return transId;
}

dsUInt32 dsVaultClient::sendFetchRefsRequest(dsInt32 nodeId)
{
    waitForStatus();

    const dsNetMsg* request = GET_Cli2Vault(kCli2Vault_VaultFetchNodeRefs);
    msgparm_t* reqBuf = AllocMessage(request);
    dsUInt32 transId = nextTransId();
    reqBuf[0] = transId;
    reqBuf[1] = nodeId;
    fSock.sendMsg(reqBuf, request);

    FreeMessage(reqBuf, request);
    return transId;
}

dsUInt32 dsVaultClient::sendSaveRequest(dsInt32 nodeId, const dsUuid& revisionId,
                                        dsUByte* nodeData, dsUInt32 nodeSize)
{
    waitForStatus();

    const dsNetMsg* request = GET_Cli2Vault(kCli2Vault_VaultNodeSave);
    msgparm_t* reqBuf = AllocMessage(request);
    dsUInt32 transId = nextTransId();
    reqBuf[0] = transId;
    reqBuf[1] = nodeId;
    memcpy((void*)(reqBuf[2]), &revisionId, sizeof(dsUuid));
    reqBuf[3] = nodeSize;
    reqBuf[4] = (msgparm_t)nodeData;
    fSock.sendMsg(reqBuf, request);

    // Don't delete twice
    reqBuf[4] = 0;
    FreeMessage(reqBuf, request);
    return transId;
}

dsUInt32 dsVaultClient::sendFindRequest(dsUByte* nodeData, dsUInt32 nodeSize)
{
    waitForStatus();

    const dsNetMsg* request = GET_Cli2Vault(kCli2Vault_VaultNodeFind);
    msgparm_t* reqBuf = AllocMessage(request);
    dsUInt32 transId = nextTransId();
    reqBuf[0] = transId;
    reqBuf[1] = nodeSize;
    reqBuf[2] = (msgparm_t)nodeData;
    fSock.sendMsg(reqBuf, request);

    // Don't delete twice
    reqBuf[2] = 0;
    FreeMessage(reqBuf, request);
    return transId;
}

dsUInt32 dsVaultClient::sendAddRequest(dsInt32 parent, dsInt32 child, dsInt32 saver)
{
    waitForStatus();

    const dsNetMsg* request = GET_Cli2Vault(kCli2Vault_VaultNodeAdd);
    msgparm_t* reqBuf = AllocMessage(request);
    dsUInt32 transId = nextTransId();
    reqBuf[0] = transId;
    reqBuf[1] = parent;
    reqBuf[2] = child;
    reqBuf[3] = saver;
    fSock.sendMsg(reqBuf, request);

    FreeMessage(reqBuf, request);
    return transId;
}
