#include "dsVaultServer.h"
#include "../dsConfigMgr.h"
#include "../dsLog.h"
#include "../Connection.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>

#include "dsVaultNode.h"

/* dsVaultComm */
void dsVaultComm::run()
{
    dsLog(kLogInfo, "Client connected");

    VaultConnectHeader connectInfo;
    fSock.peek(&connectInfo.fHeaderSize, sizeof(dsUInt32));
    if (connectInfo.fHeaderSize != sizeof(connectInfo)) {
        dsLog(kLogError, "Invalid Vault Server connection packet");
        return;
    }
    if (fSock.recv(&connectInfo, sizeof(connectInfo)) != sizeof(connectInfo)) {
        dsLog(kLogError, "Invalid Vault Server connection packet");
        return;
    }
    if (connectInfo.fToken != dsVaultServer::sVaultToken) {
        dsLogF(kLogError, "Got incorrect auth token: %s",
                          connectInfo.fToken.asString().cstr().data());
        dsLog(kLogError, "Terminating connection");
        return;
    }

    NetCli_Cli2Srv_Connect cryptConn;
    if (fSock.recv(&cryptConn.fMessage, sizeof(dsUByte)) != sizeof(dsUByte)
        || cryptConn.fMessage != kNetCliCli2SrvConnect) {
        dsLog(kLogError, "Invalid NetCli connection packet");
        return;
    }
    if (fSock.recv(&cryptConn.fLength, sizeof(dsUByte)) != sizeof(dsUByte)) {
        dsLog(kLogError, "Invalid NetCli connection packet");
        return;
    }

    size_t keyDataLen = cryptConn.fLength - (sizeof(dsUByte) * 2);
    if (keyDataLen > 64) {
        dsLog(kLogError, "Seed value too long");
        return;
    }
    memset(cryptConn.fKeyYData, 0, 64);
    fSock.recv(cryptConn.fKeyYData, keyDataLen);

    NetCli_Srv2Cli_Encrypt cryptReply;
    cryptReply.fMessage = kNetCliSrv2CliEncrypt;
    cryptReply.fLength = sizeof(cryptReply);
    if (!calculateKeys(dsBigInteger(cryptConn.fKeyYData, keyDataLen))) {
        dsLog(kLogError, "Error setting up encryption keys");
        return;
    }
    memcpy(cryptReply.fSeed, fServerSeed, 7);
    fSock.send(&cryptReply, sizeof(cryptReply));
    fSock.init(7, fSharedKey);

    dsUInt16 msgType;
    msgparm_t* replyBuf;
    while (fSock.recv(&msgType, sizeof(dsUInt16)) > 0) {
        const dsNetMsg* msg = GET_Cli2Vault(msgType);
        msgparm_t* reqBuf = fSock.recvMsg(msg);

        switch (msgType) {
        case kCli2Vault_PingRequest:
            {
                const dsNetMsg* reply = GET_Vault2Cli(kVault2Cli_PingReply);
                replyBuf = AllocMessage(reply);
                replyBuf[0] = reqBuf[0];    // Ping Time
                fSock.sendMsg(replyBuf, reply);
                FreeMessage(replyBuf, reply);
            }
            break;
        case kCli2Vault_VaultNodeCreate:
            {
                dsVaultNode node;
                node.parse((dsUByte*)(reqBuf[2]), reqBuf[1]);
                if (node.fCreatorIdx == -1)
                    node.setCreatorIdx(dsVaultServer::RootIdx());
                dsString queryStr = "INSERT INTO vault.\"Nodes\" ("
                                  + node.queryStrFields() + ") VALUES ("
                                  + node.queryStrValues(fVaultDb) + ") RETURNING idx";
                dsPostgresResult result = fVaultDb->query(queryStr);

                const dsNetMsg* reply = GET_Vault2Cli(kVault2Cli_VaultNodeCreated);
                replyBuf = AllocMessage(reply);
                replyBuf[0] = reqBuf[0];
                if (!result.isValid() || result.getNumRows() < 1) {
                    dsLogF(kLogError, "Error creating node: %s",
                           fVaultDb->getErrorMsg());
                    replyBuf[1] = kNetErrInternalError;
                    replyBuf[2] = 0;
                } else {
                    replyBuf[1] = kNetSuccess;
                    replyBuf[2] = result[0][(int)0].getInt();
                }
                fSock.sendMsg(replyBuf, reply);
                FreeMessage(replyBuf, reply);
            }
            break;
        case kCli2Vault_VaultNodeFetch:
            {
                dsString queryStr = "SELECT * FROM vault.\"Nodes\" WHERE "
                                  + dsString::Format("idx=%lld", reqBuf[1]);
                dsPostgresResult result = fVaultDb->query(queryStr);

                const dsNetMsg* reply = GET_Vault2Cli(kVault2Cli_VaultNodeFetched);
                replyBuf = AllocMessage(reply);
                replyBuf[0] = reqBuf[0];
                if (!result.isValid()) {
                    dsLogF(kLogError, "Error fetching node: %s", fVaultDb->getErrorMsg());
                    replyBuf[1] = kNetErrInternalError;
                    replyBuf[2] = 0;
                    replyBuf[3] = 0;
                } else if (result.getNumRows() == 0) {
                    dsLogF(kLogWarning, "Requested Node %lld does not exist!", reqBuf[1]);
                    replyBuf[1] = kNetErrVaultNodeNotFound;
                    replyBuf[2] = 0;
                    replyBuf[3] = 0;
                } else {
                    dsVaultNode node;
                    node.fromDbRow(result.fetchRow());
                    replyBuf[1] = kNetSuccess;
                    size_t nodeSize = node.calcSize();
                    dsUByte* nodeData = new dsUByte[nodeSize];
                    node.write(nodeData, nodeSize);
                    replyBuf[2] = nodeSize;
                    replyBuf[3] = (msgparm_t)nodeData;
                }
                fSock.sendMsg(replyBuf, reply);
                FreeMessage(replyBuf, reply);
            }
            break;
        case kCli2Vault_VaultNodeSave:
            {
                dsVaultNode node;
                node.parse((dsUByte*)(reqBuf[4]), reqBuf[3]);
                node.modifyNow();
                dsString queryStr = "UPDATE vault.\"Nodes\" SET "
                                  + node.queryStrUpdate(fVaultDb) + " WHERE "
                                  + dsString::Format("idx=%lld", reqBuf[1]);
                dsPostgresResult result = fVaultDb->query(queryStr);

                const dsNetMsg* reply = GET_Vault2Cli(kVault2Cli_VaultSaveNodeReply);
                replyBuf = AllocMessage(reply);
                replyBuf[0] = reqBuf[0];
                if (!result.isValid()) {
                    dsLogF(kLogError, "Error updating node: %s",
                           fVaultDb->getErrorMsg());
                    replyBuf[1] = kNetErrInternalError;
                } else {
                    const dsNetMsg* msg = GET_Vault2Cli(kVault2Cli_VaultNodeChanged);
                    msgparm_t* msgBuf = AllocMessage(msg);
                    msgBuf[0] = reqBuf[1];
                    memcpy((void*)(msgBuf[1]), (void*)(reqBuf[2]), sizeof(dsUuid));
                    dsVaultServer::propagateMessage(msgBuf, msg);
                    FreeMessage(msgBuf, msg);
                    replyBuf[1] = kNetSuccess;
                }
                fSock.sendMsg(replyBuf, reply);
                FreeMessage(replyBuf, reply);
            }
            break;
        //case kCli2Vault_VaultNodeDelete:
        case kCli2Vault_VaultNodeAdd:
            {
                dsString queryStr = "INSERT INTO vault.\"NodeRefs\" "
                                    "(\"ParentIdx\", \"ChildIdx\", \"SaverIdx\") VALUES "
                                  + dsString::Format("(%lld, %lld, %lld) ", reqBuf[1], reqBuf[2], reqBuf[3])
                                  + "RETURNING idx";
                dsPostgresResult result = fVaultDb->query(queryStr);

                const dsNetMsg* reply = GET_Vault2Cli(kVault2Cli_VaultAddNodeReply);
                replyBuf = AllocMessage(reply);
                replyBuf[0] = reqBuf[0];
                if (!result.isValid() || result.getNumRows() < 1) {
                    dsLogF(kLogError, "Error adding node: %s", fVaultDb->getErrorMsg());
                    replyBuf[1] = kNetErrInternalError;
                } else {
                    // Send the add notification to all clients
                    const dsNetMsg* msg = GET_Vault2Cli(kVault2Cli_VaultNodeAdded);
                    msgparm_t* msgBuf = AllocMessage(msg);
                    msgBuf[0] = reqBuf[1];
                    msgBuf[1] = reqBuf[2];
                    msgBuf[2] = reqBuf[3];
                    dsVaultServer::propagateMessage(msgBuf, msg);
                    FreeMessage(msgBuf, msg);
                    replyBuf[1] = kNetSuccess;
                }
                fSock.sendMsg(replyBuf, reply);
                FreeMessage(replyBuf, reply);
            }
            break;
        //case kCli2Vault_VaultNodeRemove:
        case kCli2Vault_VaultFetchNodeRefs:
            {
                dsString queryStr = dsString::Format("SELECT * FROM vault.fetch_tree(%lld)", reqBuf[1]);
                dsPostgresResult result = fVaultDb->query(queryStr);

                const dsNetMsg* reply = GET_Vault2Cli(kVault2Cli_VaultNodeRefsFetched);
                replyBuf = AllocMessage(reply);
                replyBuf[0] = reqBuf[0];
                if (!result.isValid()) {
                    dsLogF(kLogError, "Error fetching refs: %s", fVaultDb->getErrorMsg());
                    replyBuf[1] = kNetErrInternalError;
                    replyBuf[2] = 0;
                    replyBuf[3] = 0;
                } else {
                    replyBuf[1] = kNetSuccess;
                    size_t numRefs = result.getNumRows();
                    replyBuf[2] = numRefs;
                    if (numRefs > 0) {
                        dsVaultNodeRef* refs = new dsVaultNodeRef[numRefs];
                        for (size_t i=0; i<numRefs; i++) {
                            dsPostgresRow row = result.fetchRow();
                            refs[i].fParentIdx = row["\"ParentIdx\""].getInt();
                            refs[i].fChildIdx = row["\"ChildIdx\""].getInt();
                            refs[i].fOwnerIdx = row["\"SaverIdx\""].getInt();
                            refs[i].fSeen = row["\"SeenIdx\""].getInt() != 0;
                        }
                        replyBuf[3] = (msgparm_t)refs;
                    } else {
                        replyBuf[3] = 0;
                    }
                }
                fSock.sendMsg(replyBuf, reply);
                FreeMessage(replyBuf, reply);
            }
            break;
        case kCli2Vault_VaultNodeFind:
            {
                dsVaultNode node;
                node.parse((dsUByte*)(reqBuf[2]), reqBuf[1]);
                dsString queryStr = "SELECT idx FROM vault.\"Nodes\" WHERE "
                                  + node.queryStrSelect(fVaultDb);
                dsPostgresResult result = fVaultDb->query(queryStr);

                const dsNetMsg* reply = GET_Vault2Cli(kVault2Cli_VaultNodeFindReply);
                replyBuf = AllocMessage(reply);
                replyBuf[0] = reqBuf[0];
                if (!result.isValid()) {
                    dsLogF(kLogError, "Error finding nodes: %s",
                           fVaultDb->getErrorMsg());
                    replyBuf[1] = kNetErrInternalError;
                    replyBuf[2] = 0;
                    replyBuf[3] = 0;
                } else {
                    replyBuf[1] = kNetSuccess;
                    size_t nodeCount = result.getNumRows();
                    replyBuf[2] = nodeCount;
                    if (nodeCount > 0) {
                        dsInt32* nodeIds = new dsInt32[nodeCount];
                        for (size_t i=0; i<nodeCount; i++) {
                            dsPostgresRow row = result.fetchRow();
                            nodeIds[i] = row[(int)0].getInt();
                        }
                        replyBuf[3] = (msgparm_t)nodeIds;
                    } else {
                        replyBuf[3] = 0;
                    }
                }
                fSock.sendMsg(replyBuf, reply);
                FreeMessage(replyBuf, reply);
            }
            break;
        //case kCli2Vault_VaultSetSeen:
        //case kCli2Vault_VaultSendNode:
        default:
            dsLogF(kLogError, "Got unsupported message type: %u", msgType);
            // Continue anyway
        }

        FreeMessage(reqBuf, msg);
    }

    fSock.flush();
    dsLog(kLogInfo, "Client disconnected");
}

bool dsVaultComm::calculateKeys(const dsBigInteger& Y)
{
    dsString privateKeyPath = dsConfigMgr::GetPath("VaultPrivateKey");
    dsString publicKeyPath = dsConfigMgr::GetPath("VaultPublicKey");

    // Server's xor key for client
    dsUByte tempKey[] = { 0x39, 0xA4, 0x2C, 0xD3, 0x91, 0x1E, 0x08 };
    memcpy(fServerSeed, tempKey, 7);

    // Server communication encryption
    dsUByte k_seed[64];
    FILE* keyF = fopen(privateKeyPath.cstr(), "rb");
    if (keyF == NULL)
        return false;
    fread(k_seed, sizeof(dsUByte), 64, keyF);
    fclose(keyF);

    dsUByte n_seed[64];
    keyF = fopen(publicKeyPath.cstr(), "rb");
    if (keyF == NULL)
        return false;
    fread(n_seed, sizeof(dsUByte), 64, keyF);
    fclose(keyF);

    dsBigInteger K(k_seed, 64);
    dsBigInteger N(n_seed, 64);
    dsBigInteger clientSeed = Y.PowMod(K, N);

    dsUByte* seedData;
    size_t seedSize;
    clientSeed.getData(seedData, seedSize);
    if (seedSize > 7)
        seedSize = 7;
    for (size_t i=0; i<seedSize; i++)
        fSharedKey[i] = seedData[i] ^ fServerSeed[i];
    for (size_t i=seedSize; i<7; i++)
        fSharedKey[i] = fServerSeed[i];
    delete[] seedData;

    return true;
}


/* dsVaultServer */
std::list<dsVaultComm*> dsVaultServer::fClients;
dsMutex dsVaultServer::fClientListMutex;
dsInt32 dsVaultServer::sRootIdx = 0;

dsUuid dsVaultServer::sVaultToken;
int dsVaultServer::sVaultPipe = 0;

dsUuid dsVaultServer::getAuthToken()
{
    dsString tokString = dsConfigMgr::GetString("VaultToken", NULL_TOKEN_STR);
    return dsUuid::fromString(tokString);
}

pid_t dsVaultServer::Create()
{
    sVaultToken = getAuthToken();
    if (sVaultToken == dsUuid()) {
        // Invalid or uninitialized auth token
        fprintf(stderr, "Invalid Vault auth token.  Check your dirtsand.conf\n");
        exit(1);
    }

    int pipefd[2];
    if (socketpair(AF_UNIX, SOCK_STREAM, 0, pipefd) < 0) {
        dsLogF(kLogError, "Error creating vault pipe: %s", strerror(errno));
        exit(1);
    }

    pid_t pid = fork();
    if (pid < 0) {
        dsLogF(kLogError, "Error spawning Vault server: %s", strerror(errno));
        close(pipefd[0]);
        close(pipefd[1]);
        exit(1);
    } else if (pid == 0) {
        s_listenSock.close();
        s_children.clear();
        dsString logpath = dsConfigMgr::GetPath("LogDir", "VaultServer");
        dsLogProcessor* vaultLog = dsLogProcessor::Create(logpath.cstr(),
                                   (ELogType)dsConfigMgr::GetUInt("LogLevel", LOGLEVEL_DEFAULT));
        dsLogProcessor::setLog(vaultLog, true);

        dsVaultServer server(pipefd[0]);
        close(pipefd[1]);   // Child only uses read end
        server.run();
        exit(0);
    } else {
        close(pipefd[0]);   // Parent only uses write end
        sVaultPipe = pipefd[1];
        return pid;
    }
}

void dsVaultServer::SendConnection(dsSocket* sock)
{
    msghdr msg;
    iovec iov[1];
    char ccmsg[CMSG_SPACE(sizeof(int))];

    msg.msg_control = ccmsg;
    msg.msg_controllen = sizeof(ccmsg);
    msg.msg_name = NULL;
    msg.msg_namelen = 0;

    cmsghdr* cmptr = CMSG_FIRSTHDR(&msg);
    cmptr->cmsg_len = CMSG_LEN(sizeof(int));
    cmptr->cmsg_level = SOL_SOCKET;
    cmptr->cmsg_type = SCM_RIGHTS;
    *(int*)CMSG_DATA(cmptr) = (sock == NULL) ? 0 : sock->getHandle();

    // Stupidity: Needs at least one byte to send...  >.>
    iov[0].iov_base = const_cast<char*>("x");
    iov[0].iov_len = 1;
    msg.msg_iov = iov;
    msg.msg_iovlen = 1;

    if (sendmsg(sVaultPipe, &msg, 0) < 0)
        dsLogF(kLogError, "Error sending FD: %s", strerror(errno));
    if (sock != NULL)
        delete sock;
}

int dsVaultServer::getClientConnection()
{
    msghdr msg;
    iovec iov[1];
    char ccmsg[CMSG_SPACE(sizeof(int))];
    char buf[1];

    msg.msg_control = ccmsg;
    msg.msg_controllen = sizeof(ccmsg);
    msg.msg_name = NULL;
    msg.msg_namelen = 0;

    // For getting the dummy byte
    iov[0].iov_base = buf;
    iov[0].iov_len = 1;
    msg.msg_iov = iov;
    msg.msg_iovlen = 1;

    if (recvmsg(fListenPipe, &msg, 0) < 0) {
        dsLogF(kLogError, "Got invalid FD message: %s", strerror(errno));
        return -1;
    }

    cmsghdr* cmptr = CMSG_FIRSTHDR(&msg);
    if (cmptr->cmsg_type != SCM_RIGHTS) {
        dsLogF(kLogError, "Got invalid FD message type: %d", cmptr->cmsg_type);
        return -1;
    }
    return *(int*)CMSG_DATA(cmptr);
}

dsInt32 dsVaultServer::InitVault()
{
    dsVaultNode node;
    node.setNodeType(kNodeVaultServer);

    dsString queryStr = "SELECT idx FROM vault.\"Nodes\" WHERE "
                        + node.queryStrSelect(&fVaultDb);
    dsPostgresResult result = fVaultDb.query(queryStr);
    if (!result.isValid()) {
        dsLogF(kLogError, "Error initializing the vault: %s", fVaultDb.getErrorMsg());
        return 0;
    }

    if (result.getNumRows() < 1) {
        // Set up core vault nodes
        node.setTimeNow();
        queryStr = "INSERT INTO vault.\"Nodes\" ("
                 + node.queryStrFields() + ") VALUES ("
                 + node.queryStrValues(&fVaultDb) + ") RETURNING idx";
        result = fVaultDb.query(queryStr);
        if (!result.isValid() || result.getNumRows() < 1) {
            dsLogF(kLogError, "Error initializing the vault: %s", fVaultDb.getErrorMsg());
            return 0;
        }
        sRootIdx = result[0][(int)0].getInt();

        node.setCreatorIdx(sRootIdx);
        node.setNodeType(kNodeSystem);
        node.setTimeNow();
        queryStr = "INSERT INTO vault.\"Nodes\" ("
                 + node.queryStrFields() + ") VALUES ("
                 + node.queryStrValues(&fVaultDb) + ") RETURNING idx";
        result = fVaultDb.query(queryStr);
        if (!result.isValid() || result.getNumRows() < 1) {
            dsLogF(kLogError, "Error initializing the vault: %s", fVaultDb.getErrorMsg());
            return 0;
        }
        dsInt32 sysNode = result[0][(int)0].getInt();

        /*
        node.setCreatorIdx(sRootIdx);
        node.setNodeType(kNodePlayerInfoList);
        node.setInt32(0, kAllPlayersFolder);
        node.setTimeNow();
        queryStr = "INSERT INTO vault.\"Nodes\" ("
                 + node.queryStrFields() + ") VALUES ("
                 + node.queryStrValues(&fVaultDb) + ") RETURNING idx";
        result = fVaultDb.query(queryStr);
        if (!result.isValid() || result.getNumRows() < 1) {
            dsLogF(kLogError, "Error initializing the vault: %s", fVaultDb.getErrorMsg());
            return 0;
        }
        */

        node.setCreatorIdx(sRootIdx);
        node.setNodeType(kNodeFolder);
        node.setInt32(0, kAllAgeGlobalSDLNodesFolder);
        node.setTimeNow();
        queryStr = "INSERT INTO vault.\"Nodes\" ("
                 + node.queryStrFields() + ") VALUES ("
                 + node.queryStrValues(&fVaultDb) + ") RETURNING idx";
        result = fVaultDb.query(queryStr);
        if (!result.isValid() || result.getNumRows() < 1) {
            dsLogF(kLogError, "Error initializing the vault: %s", fVaultDb.getErrorMsg());
            return 0;
        }

        node.setCreatorIdx(sRootIdx);
        node.setNodeType(kNodeFolder);
        node.setInt32(0, kPublicAgesFolder);
        node.setTimeNow();
        queryStr = "INSERT INTO vault.\"Nodes\" ("
                 + node.queryStrFields() + ") VALUES ("
                 + node.queryStrValues(&fVaultDb) + ") RETURNING idx";
        result = fVaultDb.query(queryStr);
        if (!result.isValid() || result.getNumRows() < 1) {
            dsLogF(kLogError, "Error initializing the vault: %s", fVaultDb.getErrorMsg());
            return 0;
        }

        node.setCreatorIdx(sRootIdx);
        node.setNodeType(kNodeFolder);
        node.setInt32(0, kGlobalInboxFolder);
        node.setTimeNow();
        queryStr = "INSERT INTO vault.\"Nodes\" ("
                 + node.queryStrFields() + ") VALUES ("
                 + node.queryStrValues(&fVaultDb) + ") RETURNING idx";
        result = fVaultDb.query(queryStr);
        if (!result.isValid() || result.getNumRows() < 1) {
            dsLogF(kLogError, "Error initializing the vault: %s", fVaultDb.getErrorMsg());
            return 0;
        }
        dsInt32 ginboxNode = result[0][(int)0].getInt();

        queryStr = "INSERT INTO vault.\"NodeRefs\" "
                   "(\"ParentIdx\", \"ChildIdx\", \"SaverIdx\") VALUES "
                 + dsString::Format("(%d, %d, %d) ", sysNode, ginboxNode, 0)
                 + "RETURNING idx";
        result = fVaultDb.query(queryStr);
        if (!result.isValid() || result.getNumRows() < 1) {
            dsLogF(kLogError, "Error initializing the vault: %s", fVaultDb.getErrorMsg());
            return 0;
        }

        return sRootIdx;
    } else {
        return result[0][(int)0].getInt();
    }
}

void dsVaultServer::run()
{
    // Acquire the Vault connection
    dsString host = dsConfigMgr::GetString("dbhost", DBHOST_DEFAULT);
    dsString port = dsConfigMgr::GetString("dbport", DBPORT_DEFAULT);
    dsString user = dsConfigMgr::GetString("dbuser", DBUSER_DEFAULT);
    dsString pass = dsConfigMgr::GetString("dbpass", DBPASS_DEFAULT);
    dsString sslMode = dsConfigMgr::GetString("dbsslmode", DBSSLMODE_DEFAULT);
    dsString dbname = dsConfigMgr::GetString("dbname", DBNAME_DEFAULT);
    if (!fVaultDb.connect(host.cstr(), port.cstr(), user.cstr(), pass.cstr(),
                          sslMode.cstr(), dbname.cstr())) {
        dsLogF(kLogError, "Could not connect to Vault Database: %s",
                          fVaultDb.getErrorMsg());
        fVaultDb.close();
        return;
    }

    sRootIdx = InitVault();
    if (sRootIdx == 0) {
        fVaultDb.close();
        return;
    }
    dsLogF(kLogInfo, "Vault server started (%d)", sRootIdx);

    while (true) {
        int cliSock = getClientConnection();
        if (cliSock < 0)
            break;
        dsVaultComm* client = new dsVaultComm(cliSock, &fVaultDb);
        client->start();
        fClientListMutex.lock();
        fClients.push_back(client);
        fClientListMutex.unlock();
    }

    dsLog(kLogInfo, "Vault Server closed");
}

void dsVaultServer::destroyConnection(dsVaultComm* self)
{
    fClientListMutex.lock();
    std::list<dsVaultComm*>::iterator it = fClients.begin();
    while (it != fClients.end()) {
        if (*it == self) {
            it = fClients.erase(it);
            delete *it;
        } else {
            it++;
        }
    }
    fClientListMutex.unlock();
}

void dsVaultServer::propagateMessage(const msgparm_t* data, const dsNetMsg* msg)
{
    std::list<dsVaultComm*>::iterator i;
    fClientListMutex.lock();
    for (i = fClients.begin(); i != fClients.end(); i++)
        (*i)->sendMsg(data, msg);
    fClientListMutex.unlock();
}
